import glob
import logging
import os

import click
from click import UsageError

from aspire.commands import log_level_option
from aspire.noise import WhiteNoiseEstimator
from aspire.source.coordinates import BoxesCoordinateSource, CentersCoordinateSource
from aspire.utils.logging import setConsoleLoggingLevel

logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "--mrc_paths",
    required=True,
    help="Text file or glob expression containing paths to micrograph files",
)
@click.option(
    "--coord_paths",
    required=True,
    help="Text file or glob expression containing paths to coordinate files",
)
@click.option(
    "--starfile_out",
    required=True,
    help="Path to starfile of the particle stack to be created",
)
@click.option(
    "--particle_size",
    default=None,
    type=int,
    help="Desired box size (in pixels) of particles to be extracted",
)
@click.option(
    "--centers",
    is_flag=True,
    help="Set this flag if coordinate files contain (X,Y) particle centers",
)
@click.option(
    "--downsample",
    default=0,
    type=int,
    help="Downsample the images to this resolution prior to saving to starfile/.mrcs stack",
)
@click.option(
    "--normalize_bg",
    is_flag=True,
    help="Normalize the images to have mean zero and variance one in the corners",
)
@click.option(
    "--whiten",
    is_flag=True,
    help="Estimate the noise variance of the images and whiten",
)
@click.option(
    "--invert_contrast",
    is_flag=True,
    help="Invert the contrast of the images to ensure that clean particles have positive intensity",
)
@click.option(
    "--batch_size", default=512, help="Batch size to load images from .mrc files"
)
@click.option(
    "--save_mode",
    default="single",
    help="Option to save MRC file. If not single, saved to multiple files by batch size",
)
@click.option(
    "--overwrite", is_flag=True, help="Overwrite output if it already exists?"
)
@log_level_option
def extract_particles(
    mrc_paths,
    coord_paths,
    starfile_out,
    particle_size,
    centers,
    downsample,
    normalize_bg,
    whiten,
    invert_contrast,
    batch_size,
    save_mode,
    overwrite,
    loglevel,
):
    """
    Given a dataset of full micrographs and corresponding coordinate files
    containing the locations of picked particles in the .mrc, extract the
    particles into one or more .mrcs stacks and generate a STAR file.

    Example usage:

    aspire extract-particles --mrc_paths=my/data/*.mrc --coord_paths=my/data/coords/*.coord --starfile_out=my_dataset_stack.star --particle_size=256 --centers
    """
    # Set desired logging option for the command line
    setConsoleLoggingLevel(loglevel)

    # mrc_paths and coord_paths can be either paths to text files
    # listing the micrograph and coordinate file paths, or glob-type
    # expressions
    # first try interpreting them as files
    if os.path.exists(mrc_paths) and os.path.exists(coord_paths):
        with open(mrc_paths) as _mrc:
            mrc_files = _mrc.readlines()
        with open(coord_paths) as _coords:
            coord_files = _coords.readlines()
    elif glob.glob(mrc_paths) and glob.glob(coord_paths):
        mrc_files = sorted(glob.glob(mrc_paths))
        coord_files = sorted(glob.glob(coord_paths))
    else:
        raise UsageError(
            "--mrc_paths and --coord_paths must both be either filepaths or glob-type expressions"
        )

    # must have one-to-one micrographs and coordinate files
    if len(mrc_files) != len(coord_files):
        raise ValueError(
            f"Number of micrographs and coordinate files must match ({len(mrc_files)} micrographs and {len(coord_files)} coordinate files found)"
        )

    # this is the input to the CoordinateSource constructors
    files = list(zip(mrc_files, coord_files))

    if centers:
        if not particle_size:
            raise ValueError(
                "A --particle_size must be specified when loading from centers"
            )
        src = CentersCoordinateSource(
            files,
            particle_size=particle_size,
        )
    else:
        src = BoxesCoordinateSource(
            files,
            particle_size=particle_size,
        )

    # optional preprocessing steps
    if 0 < downsample < src.L:
        src = src.downsample(downsample)
    if normalize_bg:
        src = src.normalize_background()
    if whiten:
        estimator = WhiteNoiseEstimator(src)
        src = src.whiten(estimator)
    if invert_contrast:
        src = src.invert_contrast()

    # saves to .mrcs and STAR file with column "_rlnImageName"
    src.save(
        starfile_out, batch_size=batch_size, save_mode=save_mode, overwrite=overwrite
    )



import abc
import logging

import numpy as np
import numpy.ma as ma
from scipy.stats.mstats import winsorize

from aspire.image import Image

logger = logging.getLogger(__name__)


class ImageStacker(abc.ABC):
    """
    Interface for image and coefficient stacking classes.

    It is assumed that provided images are already aligned.
    Instances of `Stacker` act like functions.

    stacker = ImageStacker(...)
    stacker(images_to_stack).
    """

    def __init__(self):
        """
        Initialize ImageStacker instance.
        """

        # This variable will be used to keep track of shape.
        self._return_image_size = None

    @abc.abstractmethod
    def _call(self, stack):
        """

        Subclasses must implement this method.

        Given a 2D Numpy array performs the stacking computation
        returning a 1D array. Packing and unpacking the array is done
        by `_check_and_convert` and `_return` respectively in
        `__call__`.

        :param stack: 2D Numpy array.

        :return: 1D Numpy array.
        """

    def __call__(self, stack):
        """
        Stack the elements of `stack`.

        Admits an Image class, or an Numpy array.

        In the case of Numpy array, the data should be 2D
        where the first (slow) dimension is stack axis,
        and signal data is flattened to the last (fastest) axis.
        In this case an Numpy array is returned.
        This allows performing stacking arithmetic in an arbitrary
        basis.

        When passing `Image`, the stack_shape must be 1D.
        In this case an `Image` is returned.
        Users with multidimensional `Image` data
        may use `Image.stack_reshape` or stack slicing
        before/after stacking to ensure the correct
        aggregation is occurring.

        If multidimensional stacking would be useful,
        submit a feature request.

        :param stack: Image instance (or Numpy array).

        :return: Stacked data as Image (or Numpy array).
        """
        stack = self._check_and_convert(stack)

        return self._return(self._call(stack))

    def _check_and_convert(self, stack):
        """
        Check stack and returns consistent 2D np.array.

        :param stack: Image instance (or Numpy array).

        :return: 2D Numpy array.
        """

        # Careful, we need to reset this in case this stack is
        # different from last call.
        self._return_image_size = None

        # Flatten image data.
        if isinstance(stack, Image):
            # Store the image size for returning later.
            self._return_image_size = (1,) + stack.shape[1:]

            # Sanity check dimensions.
            if stack.stack_ndim != 1:
                raise ValueError(
                    f"`stack` shape of Image should be 1D for ImageStacking not {stack.stack_shape}."
                    "  Try Image.stack_reshape if needed."
                )
            stack = stack.asnumpy().reshape(stack.n_images, -1)

        # By this point we should always be an array with signal data
        # in the last axis.
        if not isinstance(stack, np.ndarray):
            raise ValueError("`stack` should be `Image` instance or Numpy array.")
        elif stack.ndim != 2:
            raise ValueError(
                f"`stack` numpy array shape should be 2D for ImageStacker not {stack.shape}."
                "  Try Image.stack_reshape if needed."
            )

        return stack

    def _return(self, result):
        """
        When ImageStacker has been passed an `Image`,
        this method creates an `Image` instance to be returned.

        In the case we started with a Numpy array this should pass
        through.

        :param result: Result data as Numpy Array.

        :return: Image(result) or Numpy array(result) based on initial `stack` input type.
        """
        if self._return_image_size is not None:
            result = Image(result.reshape(self._return_image_size))
        return result


class MeanImageStacker(ImageStacker):
    """
    Stack using `mean`.
    """

    def _call(self, stack):
        return stack.mean(axis=0)


class MedianImageStacker(ImageStacker):
    """
    Stack using `median`.
    """

    def _call(self, stack):
        return np.median(stack, axis=0)


class SigmaRejectionImageStacker(ImageStacker):
    """
    Stack using Sigma Rejection.

    When no outliers exist, sigma rejection is equivalent to `mean`.
    In the presence of outliers, pixels outside of `rejection_sigma`
    from the per-pixel-mean are discarded before the mean is
    performed.

    For potentially less Gaussian distributions, 'FWHM' and 'FWTM'
    methods are also provided.  These will take the mean of all values
    laying above half the maximum and tenth maximum
    respectively. Essentially this is the same wing clipping
    procedure, but the location of the clip is determined by the peak
    intensity instead of standard deviation.

    Note, in both cases, users are responsible for ensuring methods
    are called on reasonable data (in FW* cases we should be probably
    be using intensities). No corrections or pedestals are
    incorporated at this time, but could easily be added in the
    future.
    """

    _width_methods = {"FWHM": 0.5, "FWTM": 0.1}

    def __init__(self, rejection_sigma=3):
        """
        Instantiates with presribed `rejection_sigma`.

        :param rejection_sigma: Values falling outside
            `rejection_sigma` standard deviations are
            rejected. Defaults to 3.  Also accepts 'FWHM' and 'FWTM'
            corresponding to per-pixel full width at half maximum and
            tenth maximum respectively.

        """
        # Handle string `rejection_sigma`
        self._method = self._gaussian_method
        if isinstance(rejection_sigma, str):
            self.rejection_sigma = rejection_sigma.upper()
            self._method = self._width_method
            if self.rejection_sigma not in self._width_methods:
                raise ValueError(
                    f"`rejection_sigma` must be numeric or {self._width_methods.keys()}."
                )
        else:
            self.sigma = float(rejection_sigma)

    def _call(self, stack):
        """
        Dispatches rejection method based on `rejection_sigma`.
        """
        return self._method(stack)

    def _gaussian_method(self, stack):
        """
        Gaussian rejection.
        """

        # Compute the mean and standard deviations, pixel-wise.
        means = stack.mean(axis=0)
        std_devs = stack.std(axis=0)

        # Find values that lie outside per-pixel sigma deviations.
        outliers = np.abs(stack - means) > self.sigma * std_devs

        # Mask off the outliers.
        masked_stack = ma.masked_array(stack, mask=outliers)

        # Return mean without the outliers
        return masked_stack.mean(axis=0).data

    def _width_method(self, stack):
        """
        Width rejection.
        """

        # Compute per-pixel max.
        maxima = stack.max(axis=0)

        # Find the per-pixel value for clipping
        y = maxima * self._width_methods[self.rejection_sigma]

        # Compute outliers; values below clipping value.
        outliers = stack < y

        # Mask off the outliers.
        masked_stack = ma.masked_array(stack, mask=outliers)

        # Return mean withou the outliers
        return masked_stack.mean(axis=0).data


class WinsorizedImageStacker(ImageStacker):
    """
    Stack using Winsorization.

    Winsorizing is similar to SigmaRejectionImageStacker, except it
    admits a `percentile` and replaces rejected values with values at
    +/- `percentile`.
    """

    def __init__(self, percentile=0.1):
        """
        Instantiates WinsorizedImageStacker instance with prescribed
        `percentile`.

        See scipy docs for more details:
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.mstats.winsorize.html

        :param percentile: Float or tuple of floats.  Values must be
            [0,1] and represent the percentile used for trimming.  A
            tuple allows different lower and upper percentiles
            respectively. Example (0.1, 0.2) would Winsorize the lower
            10% and upper 20% of values.  In the case of a single
            value, the same value will be used for both upper and
            lower percentiles.
        """
        # Convert scalars to tuples.
        if not isinstance(percentile, tuple):
            percentile = (float(percentile),) * 2
        if (min(percentile) < 0) or (max(percentile) > 1):
            raise ValueError(f"`percentile` must be [0,1], passed {percentile}.")
        self.percentile = percentile

    def _call(self, stack):
        """
        Apply Winsorizing process the return the stack mean.
        """

        # Note, we intentionally disable `inplace` to avoid mutating
        # stack, just-in-case.
        stack = winsorize(stack, limits=self.percentile, inplace=False)

        # Return mean of Winsorized data.
        return stack.mean(axis=0).data


# The following will be blocked by ABC because they are not
# implemented yet, and are open to discussion/removal/implementation
# pending time.
class PoissonRejectionImageStacker(ImageStacker):
    """
    Stack using Poisson Rejection.
    """


class RobustChauvenetRejectionImageStacker(ImageStacker):
    """
    Stack using Robust Chauvenet Outlier Rejection.
    """



import logging

import numpy as np
from numpy.linalg import lstsq
from scipy.optimize import least_squares
from scipy.special import jn

from aspire.basis.basis_utils import lgwt, t_x_mat, t_x_mat_dot
from aspire.basis.pswf_2d import PSWFBasis2D
from aspire.nufft import nufft
from aspire.numeric import fft, xp
from aspire.utils import complex_type

logger = logging.getLogger(__name__)


class FPSWFBasis2D(PSWFBasis2D):
    """
    Define a derived class for fast Prolate Spheroidal Wave Function (PSWF) expanding 2D images

    The numerical evaluation for 2D PSWFs at arbitrary points in the unit disk is based on the fast method
    described in the papers as below:
        1) Boris Landa and Yoel Shkolnisky, "Steerable principal components for space-frequency localized images",
        SIAM J. Imag. Sci. 10, 508-534 (2017).
        2) Boris Landa and Yoel Shkolnisky, "Approximation scheme for essentially bandlimited and space-concentrated
        functions on a disk", Appl. Comput. Harmon. Anal. 43, 381-403 (2017).
        3) Yoel Shkolnisky, "Prolate spheroidal wave functions on a disc-Integration and approximation of
        two-dimensional bandlimited functions", Appl. Comput. Harmon. Anal. 22, 235-256 (2007).
    """

    def __init__(self, size, gamma_truncation=1.0, beta=1.0, dtype=np.float32):
        """
        Initialize an object for 2D prolate spheroidal wave function (PSWF) basis expansion using fast method.

        :param size: The size of the vectors for which to define the basis
            and the image resolution. May be a 2-tuple or an integer, in which case
            a square basis is assumed. Currently only square images are supported.
        :param gamma_trunc: Truncation parameter of PSWFs, between 0 and 1e6,
            which controls the length of the expansion and the approximation error.
            Smaller values (close to zero) guarantee smaller errors, yet longer
            expansions, and vice-versa. Note: Due to numerical considerations,
            do not exceed 1e6.
        :param beta: Bandlimit ratio relative to the Nyquist rate, between 0 and 1.
            In general, the bandlimit is c = beta*pi*(size[0]//2), therefore for
            the default value beta = 1 there is no oversampling assumed. This
            parameter controls the bandlimit of the PSWFs.
        :param dtype: Internal ndarray datatype.
        """
        super().__init__(size, gamma_truncation, beta, dtype=dtype)

    def _build(self):
        """
        Build internal data structures for the fast 2D PSWF method
        """
        logger.info("Expanding 2D images using fast PSWF method.")

        # initial the whole set of PSWF basis functions based on the bandlimit and eps error.
        self.bandlimit = self.beta * np.pi * self.rcut
        self.d_vec_all, self.alpha_all, self.lengths = self._init_pswf_func2d(
            self.bandlimit, eps=np.spacing(1)
        )

        # generate_the 2D grid and corresponding indices inside the disc.
        self._generate_grid()

        # precompute the basis functions in 2D grids
        self._precomp()

    def _precomp(self):
        """
        Precomute the PSWF functions on a polar Fourier 2D grid for the fast method
        """
        self._generate_samples()

        eps = np.spacing(1)
        a, b, c, d, e, f = self._generate_pswf_quad(
            4 * self.rcut, 2 * self.bandlimit, eps, eps, eps
        )
        self.pswf_radial_quad = self._evaluate_pswf2d_all(
            d, np.zeros(len(d)), self.max_ns
        )
        self.quad_rule_pts_x = a
        self.quad_rule_pts_y = b
        self.quad_rule_wts = c
        self.radial_quad_pts = d
        self.quad_rule_radial_wts = e
        self.num_angular_pts = f

        us_fft_pts = np.row_stack((self.quad_rule_pts_x, self.quad_rule_pts_y))
        us_fft_pts = self.bandlimit / self.rcut * us_fft_pts
        (
            blk_r,
            num_angular_pts,
            r_quad_indices,
            numel_for_n,
            indices_for_n,
            n_max,
        ) = self._pswf_integration_sub_routine()

        self.us_fft_pts = us_fft_pts
        self.blk_r = blk_r
        self.num_angular_pts = num_angular_pts
        self.r_quad_indices = r_quad_indices
        self.numel_for_n = numel_for_n
        self.indices_for_n = indices_for_n
        self.n_max = n_max
        self.size_x = len(self._disk_mask)

    def _evaluate_t(self, images):
        """
        Evaluate coefficient vectors in PSWF basis using the fast method.

        :param images: Image stack in the standard 2D coordinate basis.
        :return: Coefficient array in the PSWF basis.
        """
        # Construct array with zeros outside mask
        images_disk = np.zeros(images.shape, dtype=images.dtype)
        images_disk[:, self._disk_mask] = images[:, self._disk_mask]

        # Invoke nufft with the `many` plan (reuse plan/points)
        nfft_res = nufft(images_disk, self.us_fft_pts)

        # Accumulate coefficients
        coefficients = self._pswf_integration(nfft_res)

        return coefficients

    def _generate_pswf_quad(
        self, n, bandlimit, phi_approximate_error, lambda_max, epsilon
    ):
        """
        Generate Gaussian quadrature points and weights for 2D PSWF functions
        """
        radial_quad_points, radial_quad_weights = self._generate_pswf_radial_quad(
            n, bandlimit, phi_approximate_error, lambda_max
        )

        num_angular_points = (
            np.ceil(np.e * radial_quad_points * bandlimit / 2 - np.log(epsilon)).astype(
                "int"
            )
            + 1
        )

        for i in range(len(radial_quad_points)):
            ang_error_vec = np.absolute(
                jn(
                    range(1, 2 * num_angular_points[i] + 1),
                    bandlimit * radial_quad_points[i],
                )
            )

            num_angular_points[i] = self._sum_minus_cumsum_smaller_eps(
                ang_error_vec, epsilon
            )
            if num_angular_points[i] % 2 == 1:
                num_angular_points[i] += 1

        temp = 2 * np.pi / num_angular_points

        t = 2

        quad_rule_radial_weights = temp * radial_quad_points * radial_quad_weights
        quad_rule_weights = np.repeat(
            quad_rule_radial_weights, repeats=num_angular_points
        )
        quad_rule_pts_r = np.repeat(
            radial_quad_points, repeats=(num_angular_points / t).astype("int")
        )
        quad_rule_pts_theta = np.concatenate(
            [
                temp[i] * np.arange(num_angular_points[i] / t)
                for i in range(len(radial_quad_points))
            ]
        )

        pts_x = quad_rule_pts_r * np.cos(quad_rule_pts_theta)
        pts_y = quad_rule_pts_r * np.sin(quad_rule_pts_theta)

        return (
            pts_x,
            pts_y,
            quad_rule_weights,
            radial_quad_points,
            quad_rule_radial_weights,
            num_angular_points,
        )

    def _generate_pswf_radial_quad(
        self, n, bandlimit, phi_approximate_error, lambda_max
    ):
        """
        Generate Gaussian quadrature points and weights for the radical parts of 2D PSWFs
        """
        x, w = lgwt(20 * n, 0, 1, dtype=self.dtype)

        big_n = 0

        x_as_mat = x.reshape((len(x), 1))

        alpha_n, d_vec, approx_length = self.pswf_func2d(
            big_n, n, bandlimit, phi_approximate_error, x, w
        )

        cut_indices = np.where(
            bandlimit / 2 / np.pi * np.absolute(alpha_n) < lambda_max
        )[0]

        if len(cut_indices) == 0:
            k = len(alpha_n)
        else:
            k = cut_indices[0]

        if k % 2 == 0:
            k = k + 1

        range_array = np.arange(approx_length).reshape((1, approx_length))

        idx_for_quad_nodes = int((k + 1) / 2)
        num_quad_pts = idx_for_quad_nodes - 1

        phi_zeros = self._find_initial_nodes(
            x, n, bandlimit / 2, phi_approximate_error, idx_for_quad_nodes
        )

        def phi_for_quad_weights(t):
            return np.dot(
                t_x_mat_dot(t, big_n, range_array, approx_length), d_vec[:, : k - 1]
            )

        b = np.dot(w * np.sqrt(x), phi_for_quad_weights(x_as_mat))

        a = phi_for_quad_weights(
            phi_zeros.reshape((len(phi_zeros), 1))
        ).transpose() * np.sqrt(phi_zeros)
        init_quad_weights = lstsq(a, b, rcond=None)
        init_quad_weights = init_quad_weights[0]
        tolerance = np.spacing(1)

        def obj_func(quad_rule):
            q = quad_rule.reshape((len(quad_rule), 1))
            temp = np.dot(
                (
                    phi_for_quad_weights(q[:num_quad_pts]) * np.sqrt(q[:num_quad_pts])
                ).transpose(),
                q[num_quad_pts:],
            )
            temp = temp.reshape(temp.shape[0])
            return temp - b

        arr_to_send = np.concatenate((phi_zeros, init_quad_weights))
        quad_rule_final = least_squares(
            obj_func, arr_to_send, xtol=tolerance, ftol=tolerance, max_nfev=1000
        )
        quad_rule_final = quad_rule_final.x
        quad_rule_pts = quad_rule_final[:num_quad_pts]
        quad_rule_weights = quad_rule_final[num_quad_pts:]
        return quad_rule_pts, quad_rule_weights

    def _find_initial_nodes(
        self, x, n, bandlimit, phi_approximate_error, idx_for_quad_nodes
    ):
        """
        Find initial quadrature nodes
        """
        big_n = 0

        d_vec, approx_length, range_array = self._pswf_2d_minor_computations(
            big_n, n, bandlimit, phi_approximate_error
        )

        def phi_for_quad_nodes(t):
            return np.dot(
                t_x_mat(t, big_n, range_array, approx_length),
                d_vec[:, idx_for_quad_nodes - 1],
            )

        fun_vec = phi_for_quad_nodes(x)
        sign_flipping_vec = np.where(np.sign(fun_vec[:-1]) != np.sign(fun_vec[1:]))[0]
        phi_zeros = np.zeros(idx_for_quad_nodes - 1)

        tmp = phi_for_quad_nodes(x)
        for i, j in enumerate(sign_flipping_vec[: idx_for_quad_nodes - 1]):
            new_zero = x[j] - tmp[j] * (x[j + 1] - x[j]) / (tmp[j + 1] - tmp[j])
            phi_zeros[i] = new_zero

        phi_zeros = np.array(phi_zeros)
        return phi_zeros

    def _sum_minus_cumsum_smaller_eps(self, x, eps):
        y = np.cumsum(np.flipud(x))
        return len(y) - np.where(y > eps)[0][0] + 1

    def _pswf_integration_sub_routine(self):
        t = 2
        num_angular_pts = (self.num_angular_pts / t).astype("int")

        r_quad_indices = [0]
        r_quad_indices.extend(num_angular_pts)
        r_quad_indices = np.cumsum(r_quad_indices, dtype="int")

        n_max = int(max(self.ang_freqs) + 1)

        numel_for_n = np.zeros(n_max, dtype="int")
        for i in range(n_max):
            numel_for_n[i] = np.count_nonzero(self.ang_freqs == i)

        indices_for_n = [0]
        indices_for_n.extend(numel_for_n)
        indices_for_n = np.cumsum(indices_for_n, dtype="int")

        blk_r = [0] * n_max
        temp_const = self.bandlimit / (2 * np.pi * self.rcut)
        for i in range(n_max):
            blk_r[i] = (
                temp_const
                * self.pswf_radial_quad[indices_for_n[i] + np.arange(numel_for_n[i]), :]
            )

        return blk_r, num_angular_pts, r_quad_indices, numel_for_n, indices_for_n, n_max

    def _pswf_integration(self, images_nufft):
        """
        Perform integration part for rotational invariant property.
        """
        # Handle both singleton and stacks
        images_nufft = np.atleast_2d(images_nufft)
        num_images = images_nufft.shape[0]

        n_max_float = float(self.n_max) / 2
        r_n_eval_mat = np.zeros(
            (len(self.radial_quad_pts), self.n_max, num_images),
            dtype=complex_type(self.dtype),
        )

        for i in range(len(self.radial_quad_pts)):
            curr_r_mat = images_nufft[
                :,
                self.r_quad_indices[i] : self.r_quad_indices[i]
                + self.num_angular_pts[i],
            ].T
            curr_r_mat = np.concatenate((curr_r_mat, np.conj(curr_r_mat)))
            fft_plan = xp.asnumpy(fft.fft(xp.asarray(curr_r_mat), axis=0))
            angular_eval = fft_plan * self.quad_rule_radial_wts[i]

            r_n_eval_mat[i, :, :] = np.tile(
                angular_eval,
                (int(max(1, np.ceil(n_max_float / self.num_angular_pts[i]))), 1),
            )[: self.n_max, :]

        r_n_eval_mat = r_n_eval_mat.reshape(
            (len(self.radial_quad_pts) * self.n_max, num_images), order="F"
        )
        coeff_vec_quad = np.zeros(
            (num_images, len(self.ang_freqs)), dtype=complex_type(self.dtype)
        )
        m = self.pswf_radial_quad.shape[1]
        for i in range(self.n_max):
            coeff_vec_quad[
                :, self.indices_for_n[i] + np.arange(self.numel_for_n[i])
            ] = np.dot(self.blk_r[i], r_n_eval_mat[i * m : (i + 1) * m, :]).T

        return coeff_vec_quad



"""
Miscellaneous utilities for common unit conversions.
"""

import logging
import math

import numpy as np

logger = logging.getLogger(__name__)


def ratio_to_decibel(p):
    """
    Convert a ratio of powers to decibel (log) scale.

    Follows numpy broadcasting rules.

    :param p: Power ratio.
    :returns: Power ratio in log scale
    """

    return 10 * np.log10(p)


def voltage_to_wavelength(voltage):
    """
    Convert from electron voltage to wavelength.

    :param voltage: float, The electron voltage in kV.
    :return: float, The electron wavelength in angstroms.
    """
    # We use de Broglie's relativistic formula for wavelength given by:
    # wavelength = h / np.sqrt(2 * m * q * V * (1 + q * V / (2 * m * c**2))),
    # where
    # h = float(6.62607015e-34) is Planck's constant
    # q = float(1.602176634e-19) is elementary charge
    # m = float(9.1093837015e-31) is electron mass
    # c = float(299792458) is speed of light

    # We precalculate a = 1e10 * a / np.sqrt(2*m*q) and b = 1e6 * q / (2*m*c^2).
    # 1e10 and 1e6 are conversions from meters to angstroms and volts to kilovolts, respectively.
    a = float(12.264259661581491)
    b = float(0.9784755917869367)

    return a / math.sqrt(voltage * 1e3 + b * voltage**2)


def wavelength_to_voltage(wavelength):
    """
    Convert from electron voltage to wavelength.

    :param wavelength: float, The electron wavelength in angstroms.
    :return: float, The electron voltage in kV.
    """
    a = float(12.264259661581491)
    b = float(0.9784755917869367)

    return (-1e3 + math.sqrt(1e6 + 4 * a**2 * b / wavelength**2)) / (2 * b)



# We'll tell isort not to sort these base classes
# isort: off

from .basis import Basis
from .steerable import SteerableBasis2D
from .fb import FBBasisMixin

# isort: on

from .fb_2d import FBBasis2D
from .fb_3d import FBBasis3D
from .ffb_2d import FFBBasis2D
from .ffb_3d import FFBBasis3D
from .fle_2d import FLEBasis2D
from .fpswf_2d import FPSWFBasis2D
from .fpswf_3d import FPSWFBasis3D
from .fspca import FSPCABasis
from .pswf_2d import PSWFBasis2D
from .pswf_3d import PSWFBasis3D



import shutil

import pooch

from aspire import config
from aspire.downloader import file_to_method_map, registry, registry_urls
from aspire.volume import Volume

# Initialize pooch data fetcher instance.
_data_fetcher = pooch.create(
    # Set the cache path defined in the config. By default, the cache
    # folder operating system dependent, set by `pooch.os_cache`.
    # Pooch uses appdirs (https://github.com/ActiveState/appdirs) to
    # select an appropriate directory for the cache on each platform.
    path=config["common"]["cache_dir"].as_filename(),
    # The remote data is on Zenodo, `base_url` is a required param,
    # even though we override using individual urls in the registry.
    base_url="https://zenodo.org/communities/computationalcryoem/",
    registry=registry,
    urls=registry_urls,
)


def fetch_data(dataset_name):
    """
    The `fetch_data` method returns the full path to the downloaded data file.

    If it is not in the local storage, it will be downloaded. If the hash of the
    file in local storage doesn’t match the one in the registry, will download a
    new copy of the file. This is considered a sign that the file was updated in
    the remote storage. If the hash of the downloaded file still doesn’t match the
    one in the registry, will raise an exception to warn of possible file corruption.

    :param dataset_name: The file name (as appears in the registry) to
        fetch from local storage.
    :return: The absolute path (including the file name) of the file in
        local storage.
    """
    return _data_fetcher.fetch(dataset_name)


def download_all():
    """
    Download all ASPIRE example data and return a dictionary of filepaths.

    :return: A dictionary of method names and associated file paths.
    """

    file_paths = {}
    for data_set in registry:
        path = fetch_data(data_set)
        name = file_to_method_map[data_set]
        file_paths[name] = path

    return file_paths


def remove_downloads():
    """
    Remove the downloads directory.
    """
    shutil.rmtree(_data_fetcher.abspath)


def available_downloads():
    """
    List all available downloads.

    :return: A list of method names for downloadable files.
    """
    return list(file_to_method_map.values())


def emdb_2660():
    """
    Downloads the EMDB-2660 volume map and returns a `Volume` instance.

    Cryo-EM structure of the Plasmodium falciparum 80S ribosome
    bound to the anti-protozoan drug emetine.

    :return: A `Volume` instance.
    """
    file_path = fetch_data("emdb_2660.map")
    vol = Volume.load(file_path)

    return vol


def emdb_8012():
    """
    Downloads the EMDB-8012 volume map and returns the file path.

    The overall structure of the yeast spliceosomal U4/U6.U5 tri-snRNP at 3.7 Angstrom.

    :return: A 'Volume' instance.
    """
    file_path = fetch_data("emdb_8012.map")
    vol = Volume.load(file_path)

    return vol


def emdb_2984():
    """
    Downloads the EMDB-2984 volume map and returns the file path.

    2.2 A resolution cryo-EM structure of beta-galactosidase in complex with a cell-permeant inhibitor.
    This molecule exhibits D2 symmetry.

    :return: A 'Volume' instance.
    """
    file_path = fetch_data("emdb_2984.map")
    vol = Volume.load(file_path, symmetry_group="D2")

    return vol


def emdb_8511():
    """
    Downloads the EMDB-8511 volume map and returns the file path.

    Structure of the human HCN1 hyperpolarization-activated cyclic nucleotide-gated ion channel.
    This molecule exhibits C4 symmetry.

    :return: A 'Volume' instance.
    """
    file_path = fetch_data("emdb_8511.map")
    vol = Volume.load(file_path, symmetry_group="C4")

    return vol


def emdb_3645():
    """
    Downloads the EMDB-3645 volume map and returns the file path.

    CryoEM density of TcdA1 in prepore state (SPHIRE tutorial).
    This molecule exhibits C5 symmetry.

    :return: A 'Volume' instance.
    """
    file_path = fetch_data("emdb_3645.map")
    vol = Volume.load(file_path, symmetry_group="C5")

    return vol


def emdb_4905():
    """
    Downloads the EMDB-4905 volume map and returns the file path.

    3D structure of horse spleen apoferritin determined using multifunctional
    graphene supports for electron cryomicroscopy. This molecule exhibits octahedral symmetry.

    :return: A 'Volume' instance.
    """
    file_path = fetch_data("emdb_4905.map")
    vol = Volume.load(file_path, symmetry_group="O")

    return vol


def emdb_10835():
    """
    Downloads the EMDB-10835 volume map and returns the file path.

    High resolution cryo-EM structure of urease from the pathogen Yersinia enterocolitica.
    This molecule exhibits tetrahedral symmetry.

    :return: A 'Volume' instance.
    """
    file_path = fetch_data("emdb_10835.map")
    vol = Volume.load(file_path, symmetry_group="T")

    return vol


def emdb_5778():
    """
    Downloads the EMDB-5778 volume map and returns the file path.

    Structure of the capsaicin receptor, TRPV1, determined by single particle electron cryo-microscopy.
    This molecule exhibits C4 symmetry.

    :return: A 'Volume' instance.
    """
    file_path = fetch_data("emdb_5778.map")
    vol = Volume.load(file_path, symmetry_group="C4")

    return vol


def emdb_6287():
    """
    Downloads the EMDB-6287 volume map and returns the file path.

    2.8 Angstrom resolution reconstruction of the T20S proteasome.
    This molecule exhibits D7 symmetry.

    :return: A 'Volume' instance.
    """
    file_path = fetch_data("emdb_6287.map")
    vol = Volume.load(file_path, symmetry_group="D7")

    return vol


def emdb_2824():
    """
    Downloads the EMDB-2824 volume map and returns the file path.

    Beta-galactosidase reconstruction.
    This molecule exhibits C2 symmetry.

    :return: A 'Volume' instance.
    """
    file_path = fetch_data("emdb_2824.map")
    vol = Volume.load(file_path, symmetry_group="C2")

    return vol


def emdb_14621():
    """
    Downloads the EMDB-14621 volume map and returns the file path.

    Map of SARSCoV2 spike protein.
    This molecule exhibits C3 symmetry.

    :return: A 'Volume' instance.
    """
    file_path = fetch_data("emdb_14621.map")
    vol = Volume.load(file_path, symmetry_group="C3")

    return vol


def emdb_2484():
    """
    Downloads the EMDB-2484 volume map and returns the file path.

    Pre-fusion structure of trimeric HIV-1 envelope glycoprotein determined by cryo-electron microscopy.
    This molecule exhibits C3 symmetry.

    :return: A 'Volume' instance.
    """
    file_path = fetch_data("emdb_2484.map")
    vol = Volume.load(file_path, symmetry_group="C3")

    return vol


def emdb_6458():
    """
    Downloads the EMDB-6458 volume map and returns the file path.

    Cryo-EM Structure of the Activated NAIP2/NLRC4 Inflammasome Reveals Nucleated Polymerization.
    This molecule exhibits C11 symmetry.

    :return: A 'Volume' instance.
    """
    file_path = fetch_data("emdb_6458.map")
    vol = Volume.load(file_path, symmetry_group="C11")

    return vol



from .blk_diag_matrix import BlkDiagMatrix, is_scalar_type
from .diag_matrix import DiagMatrix
from .filters import (
    ArrayFilter,
    BlueFilter,
    CTFFilter,
    DualFilter,
    Filter,
    FunctionFilter,
    IdentityFilter,
    LambdaFilter,
    MultiplicativeFilter,
    PinkFilter,
    PowerFilter,
    RadialCTFFilter,
    ScalarFilter,
    ScaledFilter,
    ZeroFilter,
    evaluate_src_filters_on_grid,
)
from .polar_ft import PolarFT
from .wemd import wemd_embed, wemd_norm



import logging
from abc import ABC, abstractmethod, abstractproperty

import numpy as np

from aspire.utils import Rotation

logger = logging.getLogger(__name__)


class SymmetryGroup(ABC):
    """
    Base class for symmetry groups.
    """

    def __init__(self, dtype):
        """
        :param dtype: Numpy dtype to be used for rotation matrices.
        """
        self.dtype = np.dtype(dtype)
        self.rotations = self.generate_rotations()

    @abstractmethod
    def generate_rotations(self):
        """
        Method for generating a Rotation object for the symmetry group.
        """

    @property
    def matrices(self):
        return self.rotations.matrices

    def __repr__(self):
        return f"{self.__class__.__name__}, {self.__dict__}"

    @abstractproperty
    def to_string(self):
        """String denoting the symmetry type."""

    def __str__(self):
        return f"{self.to_string}"

    @staticmethod
    def from_string(symmetry, dtype):
        """
        Takes a string, ie. 'C1', 'C7', 'D3', 'T', 'O', and returns a concrete
        SymmetryGroup object.

        :param symmetry: A string indicating the symmetry of a molecule.
        :param dtype: dtype for rotation matrices.
        :return: Concrete SymmetryGroup object.
        """

        symmetry = symmetry.upper()
        symmetry_type = symmetry[0]
        symmetric_order = symmetry[1:]

        map_to_sym_group = {
            "C": CnSymmetryGroup,
            "D": DnSymmetryGroup,
            "T": TSymmetryGroup,
            "O": OSymmetryGroup,
        }
        if symmetry_type not in map_to_sym_group.keys():
            raise ValueError(
                f"Symmetry type {symmetry_type} not supported. Try: {*map_to_sym_group.keys(),}."
            )

        symmetry_group = map_to_sym_group[symmetry_type]
        group_kwargs = dict(dtype=dtype)
        if symmetric_order:
            group_kwargs["order"] = int(symmetric_order)

        return symmetry_group(**group_kwargs)


class CnSymmetryGroup(SymmetryGroup):
    """
    Cyclic symmetry group.
    """

    def __init__(self, order, dtype):
        """
        `CnSymmetryGroup` instance that serves up a `Rotation` object
        containing rotation matrices of the symmetry group (including
        the identity) accessed via the `matrices` attribute.

        :param order: The cyclic order for the symmetry group (int).
        :param dtype: Numpy dtype to be used for rotation matrices.
        """

        self.order = int(order)
        super().__init__(dtype=dtype)

    @property
    def to_string(self):
        return "C" + str(self.order)

    def generate_rotations(self):
        """
        The Cn symmetry group contains all rotations about the z-axis
        by multiples of 2pi/n.

        In the case of an AsymmetricVolume or LegacyVolume, `symmetry_group`
        contains only the identity.

        :return: Rotation object containing the Cn symmetry group and the identity.
        """
        angles = 2 * np.pi * np.arange(self.order) / self.order
        return Rotation.about_axis("z", angles, dtype=self.dtype)


class IdentitySymmetryGroup(CnSymmetryGroup):
    """
    The identity symmetry group.
    """

    def __init__(self, dtype):
        """
        `IdentitySymmetryGroup` instance that serves up a `Rotation` object
        containing the identity matrix.

        :param dtype: Numpy dtype to be used for rotation matrices.
        """

        super().__init__(order=1, dtype=dtype)


class DnSymmetryGroup(SymmetryGroup):
    """
    Dihedral symmetry group.
    """

    def __init__(self, order, dtype):
        """
        `DnSymmetryGroup` instance that serves up a `Rotation` object
        containing rotation matrices of the symmetry group (including
        the Identity) accessed via the `matrices` attribute. Note, this
        is the chiral dihedral symmetry group which does contain reflections.

        :param order: The cyclic order for the symmetry group (int).
        :param dtype: Numpy dtype to be used for rotation matrices.
        """

        self.order = int(order)
        super().__init__(dtype=dtype)

    @property
    def to_string(self):
        return "D" + str(self.order)

    def generate_rotations(self):
        """
        The Dn symmetry group contains all elements of the Cn symmetry group.
        In addition, for each element of the Cn symmetric group we rotate by
        pi about a perpendicular axis, in this case the y-axis.

        :return: Rotation object containing the Dn symmetry group and the identity.
        """
        # Rotations to induce cyclic symmetry
        angles = 2 * np.pi * np.arange(self.order, dtype=self.dtype) / self.order
        rot_z = Rotation.about_axis("z", angles).matrices

        # Perpendicular rotation to induce dihedral symmetry
        rot_perp = Rotation.about_axis("y", np.pi, dtype=self.dtype).matrices

        # Full set of rotations.
        rots = np.concatenate((rot_z, rot_z @ rot_perp[0]), dtype=self.dtype)

        return Rotation(rots)


class TSymmetryGroup(SymmetryGroup):
    """
    Tetrahedral symmetry group.
    """

    def __init__(self, dtype):
        """
        `TSymmetryGroup` instance that serves up a `Rotation` object
        containing rotation matrices of the symmetry group (including the
        Identity) accessed via the `matrices` attribute. Note, this is the
        chiral tetrahedral symmetry group which does not contain reflections.

        :param dtype: Numpy dtype to be used for rotation matrices.
        """

        super().__init__(dtype=dtype)

    @property
    def to_string(self):
        return "T"

    def generate_rotations(self):
        """
        A tetrahedron has C3 symmetry along the 4 axes through each vertex and
        perpendicular to the opposite face, and C2 symmetry along the axes through
        the midpoints of opposite edges. We convert from axis-angle representation of
        the symmetry group elements into rotation vectors to generate the rotation
        matrices via the `from_rotvec()` method.

        :return: Rotation object containing the tetrahedral symmetry group and the identity.
        """
        # C3 rotation vectors, ie. angle * axis.
        axes_C3 = np.array(
            [[1, 1, 1], [-1, -1, 1], [1, -1, -1], [-1, 1, -1]], dtype=self.dtype
        )
        axes_C3 /= np.linalg.norm(axes_C3, axis=-1)[..., np.newaxis]
        angles_C3 = np.array([2 * np.pi / 3, 4 * np.pi / 3], dtype=self.dtype)
        rot_vecs_C3 = np.concatenate(
            [angle * axes_C3 for angle in angles_C3], dtype=self.dtype
        )

        # C2 rotation vectors.
        axes_C2 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=self.dtype)
        rot_vecs_C2 = np.pi * axes_C2

        # The full set of rotation vectors inducing tetrahedral symmetry.
        rot_vec_I = np.zeros((1, 3), dtype=self.dtype)
        rot_vecs = np.concatenate(
            (rot_vec_I, rot_vecs_C3, rot_vecs_C2), dtype=self.dtype
        )

        # Return rotations.
        return Rotation.from_rotvec(rot_vecs, dtype=self.dtype)


class OSymmetryGroup(SymmetryGroup):
    """
    Octahedral symmetry group.
    """

    def __init__(self, dtype):
        """
        `OSymmetryGroup` instance that serves up a `Rotation` object
        containing rotation matrices of the symmetry group (including the
        Identity) accessed via the `matrices` attribute. Note, this is the
        chiral octahedral symmetry group which does not contain reflections.

        :param dtype: Numpy dtype to be used for rotation matrices.
        """
        super().__init__(dtype=dtype)

        self._symmetry_group = self.generate_rotations()

    @property
    def to_string(self):
        return "O"

    def generate_rotations(self):
        """
        The symmetry group elements of the octahedral symmetry group O are the identity,
        the elements of 3 C4 rotation groups whose axes pass through two opposite vertices of the
        regular octahedron, 4 C3 rotation groups whose axes pass through the midpoints of two of
        its opposite faces, and 6 C2 rotation groups whose axes pass through the midpoints of two
        of its opposite edges.

        :return: Rotation object containing the octahedral symmetry group and the identity.
        """

        # C4 rotation vectors, ie angle * axis
        axes_C4 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=self.dtype)
        angles_C4 = np.array([np.pi / 2, np.pi, 3 * np.pi / 2], dtype=self.dtype)
        rot_vecs_C4 = np.array(
            [angle * axes_C4 for angle in angles_C4], dtype=self.dtype
        ).reshape((9, 3))

        # C3 rotation vectors
        axes_C3 = np.array(
            [[1, 1, 1], [-1, 1, 1], [1, -1, 1], [1, 1, -1]], dtype=self.dtype
        )
        axes_C3 /= np.linalg.norm(axes_C3, axis=-1)[..., np.newaxis]
        angles_C3 = np.array([2 * np.pi / 3, 4 * np.pi / 3], dtype=self.dtype)
        rot_vecs_C3 = np.array(
            [angle * axes_C3 for angle in angles_C3], dtype=self.dtype
        ).reshape((8, 3))

        # C2 rotation vectors
        axes_C2 = np.array(
            [[1, 1, 0], [-1, 1, 0], [1, 0, 1], [-1, 0, 1], [0, 1, 1], [0, -1, 1]],
            dtype=self.dtype,
        )
        axes_C2 /= np.linalg.norm(axes_C2, axis=-1)[..., np.newaxis]
        rot_vecs_C2 = np.pi * axes_C2

        # The full set of rotation vectors inducing octahedral symmetry.
        rot_vec_I = np.zeros((1, 3), dtype=self.dtype)
        rot_vecs = np.concatenate(
            (rot_vec_I, rot_vecs_C4, rot_vecs_C3, rot_vecs_C2), dtype=self.dtype
        )

        # Return rotations.
        return Rotation.from_rotvec(rot_vecs, dtype=self.dtype)



"""Utilities for input validation"""

# Authors: Olivier Grisel
#          Gael Varoquaux
#          Andreas Mueller
#          Lars Buitinck
#          Alexandre Gramfort
#          Nicolas Tresegnie
#          Sylvain Marie
#          Garrett Wright (very minor mod for complex).
# License: BSD 3 clause

import numbers
import warnings

import numpy as np
import scipy.sparse as sp
from numpy.core.numeric import ComplexWarning
from sklearn.exceptions import DataConversionWarning
from sklearn.utils.validation import _assert_all_finite

FLOAT_DTYPES = (np.float64, np.float32, np.float16)


def _num_samples(x):
    """Return number of samples in array-like x."""
    message = "Expected sequence or array-like, got %s" % type(x)
    if hasattr(x, "fit") and callable(x.fit):
        # Don't get num_samples from an ensembles length!
        raise TypeError(message)

    if not hasattr(x, "__len__") and not hasattr(x, "shape"):
        if hasattr(x, "__array__"):
            x = np.asarray(x)
        else:
            raise TypeError(message)

    if hasattr(x, "shape") and x.shape is not None:
        if len(x.shape) == 0:
            raise TypeError(
                "Singleton array %r cannot be considered" " a valid collection." % x
            )
        # Check that shape is returning an integer or default to len
        # Dask dataframes may not return numeric shape[0] value
        if isinstance(x.shape[0], numbers.Integral):
            return x.shape[0]

    try:
        return len(x)
    except TypeError:
        raise TypeError(message)


def _ensure_sparse_format(
    spmatrix, accept_sparse, dtype, copy, force_all_finite, accept_large_sparse
):
    """Convert a sparse matrix to a given format.

    Checks the sparse format of spmatrix and converts if necessary.

    Parameters
    ----------
    spmatrix : scipy sparse matrix
        Input to validate and convert.

    accept_sparse : string, boolean or list/tuple of strings
        String[s] representing allowed sparse matrix formats ('csc',
        'csr', 'coo', 'dok', 'bsr', 'lil', 'dia'). If the input is sparse but
        not in the allowed format, it will be converted to the first listed
        format. True allows the input to be any format. False means
        that a sparse matrix input will raise an error.

    dtype : string, type or None
        Data type of result. If None, the dtype of the input is preserved.

    copy : boolean
        Whether a forced copy will be triggered. If copy=False, a copy might
        be triggered by a conversion.

    force_all_finite : boolean or 'allow-nan', (default=True)
        Whether to raise an error on np.inf and np.nan in X. The possibilities
        are:

        - True: Force all values of X to be finite.
        - False: accept both np.inf and np.nan in X.
        - 'allow-nan': accept only np.nan values in X. Values cannot be
          infinite.

        .. versionadded:: 0.20
           ``force_all_finite`` accepts the string ``'allow-nan'``.

    Returns
    -------
    spmatrix_converted : scipy sparse matrix.
        Matrix that is ensured to have an allowed type.
    """
    if dtype is None:
        dtype = spmatrix.dtype

    changed_format = False

    if isinstance(accept_sparse, str):
        accept_sparse = [accept_sparse]

    # Indices dtype validation
    _check_large_sparse(spmatrix, accept_large_sparse)

    if accept_sparse is False:
        raise TypeError(
            "A sparse matrix was passed, but dense "
            "data is required. Use X.toarray() to "
            "convert to a dense numpy array."
        )
    elif isinstance(accept_sparse, (list, tuple)):
        if len(accept_sparse) == 0:
            raise ValueError(
                "When providing 'accept_sparse' "
                "as a tuple or list, it must contain at "
                "least one string value."
            )
        # ensure correct sparse format
        if spmatrix.format not in accept_sparse:
            # create new with correct sparse
            spmatrix = spmatrix.asformat(accept_sparse[0])
            changed_format = True
    elif accept_sparse is not True:
        # any other type
        raise ValueError(
            "Parameter 'accept_sparse' should be a string, "
            "boolean or list of strings. You provided "
            "'accept_sparse={}'.".format(accept_sparse)
        )

    if dtype != spmatrix.dtype:
        # convert dtype
        spmatrix = spmatrix.astype(dtype)
    elif copy and not changed_format:
        # force copy
        spmatrix = spmatrix.copy()

    if force_all_finite:
        if not hasattr(spmatrix, "data"):
            warnings.warn(
                "Can't check %s sparse matrix for nan or inf." % spmatrix.format,
                stacklevel=2,
            )
        else:
            _assert_all_finite(spmatrix.data, allow_nan=force_all_finite == "allow-nan")

    return spmatrix


def _ensure_no_complex_data(array):
    if (
        hasattr(array, "dtype")
        and array.dtype is not None
        and hasattr(array.dtype, "kind")
        and array.dtype.kind == "c"
    ):
        raise ValueError("Complex data not supported\n" "{}\n".format(array))


def check_array(
    array,
    accept_sparse=False,
    accept_large_sparse=True,
    dtype="numeric",
    order=None,
    copy=False,
    force_all_finite=True,
    ensure_2d=True,
    allow_nd=False,
    ensure_min_samples=1,
    ensure_min_features=1,
    warn_on_dtype=None,
    estimator=None,
    allow_complex=False,
):
    """Input validation on an array, list, sparse matrix or similar.

    By default, the input is checked to be a non-empty 2D array containing
    only finite values. If the dtype of the array is object, attempt
    converting to float, raising on failure.

    Parameters
    ----------
    array : object
        Input object to check / convert.

    accept_sparse : string, boolean or list/tuple of strings (default=False)
        String[s] representing allowed sparse matrix formats, such as 'csc',
        'csr', etc. If the input is sparse but not in the allowed format,
        it will be converted to the first listed format. True allows the input
        to be any format. False means that a sparse matrix input will
        raise an error.

    accept_large_sparse : bool (default=True)
        If a CSR, CSC, COO or BSR sparse matrix is supplied and accepted by
        accept_sparse, accept_large_sparse=False will cause it to be accepted
        only if its indices are stored with a 32-bit dtype.

        .. versionadded:: 0.20

    dtype : string, type, list of types or None (default="numeric")
        Data type of result. If None, the dtype of the input is preserved.
        If "numeric", dtype is preserved unless array.dtype is object.
        If dtype is a list of types, conversion on the first type is only
        performed if the dtype of the input is not in the list.

    order : 'F', 'C' or None (default=None)
        Whether an array will be forced to be fortran or c-style.
        When order is None (default), then if copy=False, nothing is ensured
        about the memory layout of the output array; otherwise (copy=True)
        the memory layout of the returned array is kept as close as possible
        to the original array.

    copy : boolean (default=False)
        Whether a forced copy will be triggered. If copy=False, a copy might
        be triggered by a conversion.

    force_all_finite : boolean or 'allow-nan', (default=True)
        Whether to raise an error on np.inf and np.nan in array. The
        possibilities are:

        - True: Force all values of array to be finite.
        - False: accept both np.inf and np.nan in array.
        - 'allow-nan': accept only np.nan values in array. Values cannot
          be infinite.

        For object dtyped data, only np.nan is checked and not np.inf.

        .. versionadded:: 0.20
           ``force_all_finite`` accepts the string ``'allow-nan'``.

    ensure_2d : boolean (default=True)
        Whether to raise a value error if array is not 2D.

    allow_nd : boolean (default=False)
        Whether to allow array.ndim > 2.

    ensure_min_samples : int (default=1)
        Make sure that the array has a minimum number of samples in its first
        axis (rows for a 2D array). Setting to 0 disables this check.

    ensure_min_features : int (default=1)
        Make sure that the 2D array has some minimum number of features
        (columns). The default value of 1 rejects empty datasets.
        This check is only enforced when the input data has effectively 2
        dimensions or is originally 1D and ``ensure_2d`` is True. Setting to 0
        disables this check.

    warn_on_dtype : boolean or None, optional (default=None)
        Raise DataConversionWarning if the dtype of the input data structure
        does not match the requested dtype, causing a memory copy.

        .. deprecated:: 0.21
            ``warn_on_dtype`` is deprecated in version 0.21 and will be
            removed in 0.23.

    estimator : str or estimator instance (default=None)
        If passed, include the name of the estimator in warning messages.

    Returns
    -------
    array_converted : object
        The converted and validated array.
    """
    # warn_on_dtype deprecation
    if warn_on_dtype is not None:
        warnings.warn(
            "'warn_on_dtype' is deprecated in version 0.21 and will be "
            "removed in 0.23. Don't set `warn_on_dtype` to remove this "
            "warning.",
            FutureWarning,
            stacklevel=2,
        )

    # store reference to original array to check if copy is needed when
    # function returns
    array_orig = array

    # store whether originally we wanted numeric dtype
    dtype_numeric = isinstance(dtype, str) and dtype == "numeric"

    dtype_orig = getattr(array, "dtype", None)
    if not hasattr(dtype_orig, "kind"):
        # not a data type (e.g. a column named dtype in a pandas DataFrame)
        dtype_orig = None

    # check if the object contains several dtypes (typically a pandas
    # DataFrame), and store them. If not, store None.
    dtypes_orig = None
    if hasattr(array, "dtypes") and hasattr(array.dtypes, "__array__"):
        dtypes_orig = list(array.dtypes)
        # pandas boolean dtype __array__ interface coerces bools to objects
        for i, dtype_iter in enumerate(dtypes_orig):
            if dtype_iter.kind == "b":
                dtypes_orig[i] = np.object

        if all(isinstance(dtype, np.dtype) for dtype in dtypes_orig):
            dtype_orig = np.result_type(*dtypes_orig)

    if dtype_numeric:
        if dtype_orig is not None and dtype_orig.kind == "O":
            # if input is object, convert to float.
            dtype = np.float64
        else:
            dtype = None

    if isinstance(dtype, (list, tuple)):
        if dtype_orig is not None and dtype_orig in dtype:
            # no dtype conversion required
            dtype = None
        else:
            # dtype conversion required. Let's select the first element of the
            # list of accepted types.
            dtype = dtype[0]

    if force_all_finite not in (True, False, "allow-nan"):
        raise ValueError(
            'force_all_finite should be a bool or "allow-nan"'
            ". Got {!r} instead".format(force_all_finite)
        )

    if estimator is not None:
        if isinstance(estimator, str):
            estimator_name = estimator
        else:
            estimator_name = estimator.__class__.__name__
    else:
        estimator_name = "Estimator"
    context = " by %s" % estimator_name if estimator is not None else ""

    if sp.issparse(array):
        if not allow_complex:
            _ensure_no_complex_data(array)
        array = _ensure_sparse_format(
            array,
            accept_sparse=accept_sparse,
            dtype=dtype,
            copy=copy,
            force_all_finite=force_all_finite,
            accept_large_sparse=accept_large_sparse,
        )
    else:
        # If np.array(..) gives ComplexWarning, then we convert the warning
        # to an error. This is needed because specifying a non complex
        # dtype to the function converts complex to real dtype,
        # thereby passing the test made in the lines following the scope
        # of warnings context manager.
        with warnings.catch_warnings():
            try:
                warnings.simplefilter("error", ComplexWarning)
                if dtype is not None and np.dtype(dtype).kind in "iu":
                    # Conversion float -> int should not contain NaN or
                    # inf (numpy#14412). We cannot use casting='safe' because
                    # then conversion float -> int would be disallowed.
                    array = np.asarray(array, order=order)
                    if array.dtype.kind == "f":
                        _assert_all_finite(array, allow_nan=False, msg_dtype=dtype)
                    array = array.astype(dtype, casting="unsafe", copy=False)
                else:
                    array = np.asarray(array, order=order, dtype=dtype)
            except ComplexWarning:
                raise ValueError("Complex data not supported\n" "{}\n".format(array))

        if not allow_complex:
            # It is possible that the np.array(..) gave no warning. This happens
            # when no dtype conversion happened, for example dtype = None. The
            # result is that np.array(..) produces an array of complex dtype
            # and we need to catch and raise exception for such cases.
            _ensure_no_complex_data(array)

        if ensure_2d:
            # If input is scalar raise error
            if array.ndim == 0:
                raise ValueError(
                    "Expected 2D array, got scalar array instead:\narray={}.\n"
                    "Reshape your data either using array.reshape(-1, 1) if "
                    "your data has a single feature or array.reshape(1, -1) "
                    "if it contains a single sample.".format(array)
                )
            # If input is 1D raise error
            if array.ndim == 1:
                raise ValueError(
                    "Expected 2D array, got 1D array instead:\narray={}.\n"
                    "Reshape your data either using array.reshape(-1, 1) if "
                    "your data has a single feature or array.reshape(1, -1) "
                    "if it contains a single sample.".format(array)
                )

        # in the future np.flexible dtypes will be handled like object dtypes
        if dtype_numeric and np.issubdtype(array.dtype, np.flexible):
            warnings.warn(
                "Beginning in version 0.22, arrays of bytes/strings will be "
                "converted to decimal numbers if dtype='numeric'. "
                "It is recommended that you convert the array to "
                "a float dtype before using it in scikit-learn, "
                "for example by using "
                "your_array = your_array.astype(np.float64).",
                FutureWarning,
                stacklevel=2,
            )

        # make sure we actually converted to numeric:
        if dtype_numeric and array.dtype.kind == "O":
            array = array.astype(np.float64)
        if not allow_nd and array.ndim >= 3:
            raise ValueError(
                "Found array with dim %d. %s expected <= 2."
                % (array.ndim, estimator_name)
            )

        if force_all_finite:
            _assert_all_finite(array, allow_nan=force_all_finite == "allow-nan")

    if ensure_min_samples > 0:
        n_samples = _num_samples(array)
        if n_samples < ensure_min_samples:
            raise ValueError(
                "Found array with %d sample(s) (shape=%s) while a"
                " minimum of %d is required%s."
                % (n_samples, array.shape, ensure_min_samples, context)
            )

    if ensure_min_features > 0 and array.ndim == 2:
        n_features = array.shape[1]
        if n_features < ensure_min_features:
            raise ValueError(
                "Found array with %d feature(s) (shape=%s) while"
                " a minimum of %d is required%s."
                % (n_features, array.shape, ensure_min_features, context)
            )

    if warn_on_dtype and dtype_orig is not None and array.dtype != dtype_orig:
        msg = "Data with input dtype %s was converted to %s%s." % (
            dtype_orig,
            array.dtype,
            context,
        )
        warnings.warn(msg, DataConversionWarning, stacklevel=2)

    if copy and np.may_share_memory(array, array_orig):
        array = np.array(array, dtype=dtype, order=order)

    if warn_on_dtype and dtypes_orig is not None and {array.dtype} != set(dtypes_orig):
        # if there was at the beginning some other types than the final one
        # (for instance in a DataFrame that can contain several dtypes) then
        # some data must have been converted
        msg = "Data with input dtype %s were all converted to %s%s." % (
            ", ".join(map(str, sorted(set(dtypes_orig)))),
            array.dtype,
            context,
        )
        warnings.warn(msg, DataConversionWarning, stacklevel=3)

    return array


def _check_large_sparse(X, accept_large_sparse=False):
    """Raise a ValueError if X has 64bit indices and accept_large_sparse=False"""
    if not accept_large_sparse:
        supported_indices = ["int32"]
        if X.getformat() == "coo":
            index_keys = ["col", "row"]
        elif X.getformat() in ["csr", "csc", "bsr"]:
            index_keys = ["indices", "indptr"]
        else:
            return
        for key in index_keys:
            indices_datatype = getattr(X, key).dtype
            if indices_datatype not in supported_indices:
                raise ValueError(
                    "Only sparse matrices with 32-bit integer"
                    " indices are accepted. Got %s indices." % indices_datatype
                )



from .adaptive_support import adaptive_support
from .class_avg import ClassAvgSource, DebugClassAvgSource, DefaultClassAvgSource
from .denoised_src import DenoisedImageSource
from .denoiser import Denoiser
from .denoiser_cov2d import DenoiserCov2D, src_wiener_coords



import logging

import numpy as np
from numpy.linalg import eigh, norm, svd

from aspire.abinitio import CLOrient3D, SyncVotingMixin
from aspire.operators import PolarFT
from aspire.utils import (
    J_conjugate,
    Rotation,
    all_pairs,
    all_triplets,
    anorm,
    cyclic_rotations,
    tqdm,
    trange,
)
from aspire.utils.random import randn

logger = logging.getLogger(__name__)


class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
    """
    Define a class to estimate 3D orientations using common lines methods for molecules with
    C3 and C4 cyclic symmetry.

    The related publications are listed below:
    G. Pragier and Y. Shkolnisky,
    A Common Lines Approach for Abinitio Modeling of Cyclically Symmetric Molecules,
    Inverse Problems, 35, 124005, (2019).

    Y. Shkolnisky, and A. Singer,
    Viewing Direction Estimation in Cryo-EM Using Synchronization,
    SIAM J. Imaging Sciences, 5, 1088-1110 (2012).

    A. Singer, R. R. Coifman, F. J. Sigworth, D. W. Chester, Y. Shkolnisky,
    Detecting Consistent Common Lines in Cryo-EM by Voting,
    Journal of Structural Biology, 169, 312-322 (2010).
    """

    def __init__(
        self,
        src,
        symmetry=None,
        n_rad=None,
        n_theta=None,
        max_shift=0.15,
        shift_step=1,
        epsilon=1e-3,
        max_iters=1000,
        degree_res=1,
        seed=None,
    ):
        """
        Initialize object for estimating 3D orientations for molecules with C3 and C4 symmetry.

        :param src: The source object of 2D denoised or class-averaged images with metadata
        :param symmetry: A string, 'C3' or 'C4', indicating the symmetry type.
        :param n_rad: The number of points in the radial direction
        :param n_theta: The number of points in the theta direction
        :param max_shift: Maximum range for shifts as a proportion of resolution. Default = 0.15.
        :param shift_step: Resolution of shift estimation in pixels. Default = 1 pixel.
        :param epsilon: Tolerance for the power method.
        :param max_iter: Maximum iterations for the power method.
        :param degree_res: Degree resolution for estimating in-plane rotations.
        :param seed: Optional seed for RNG.
        """

        super().__init__(
            src,
            n_rad=n_rad,
            n_theta=n_theta,
            max_shift=max_shift,
            shift_step=shift_step,
        )

        self._check_symmetry(symmetry)
        self.epsilon = epsilon
        self.max_iters = max_iters
        self.degree_res = degree_res
        self.seed = seed

    def _check_symmetry(self, symmetry):
        if symmetry is None:
            raise NotImplementedError(
                "Symmetry type not supplied. Please indicate C3 or C4 symmetry."
            )
        else:
            symmetry = symmetry.upper()
            if symmetry not in ["C3", "C4"]:
                raise NotImplementedError(
                    f"Only C3 and C4 symmetry supported. {symmetry} was supplied."
                )
            self.order = int(symmetry[1])

    def estimate_rotations(self):
        """
        Estimate rotation matrices for molecules with C3 or C4 symmetry.

        :return: Array of rotation matrices, size n_imgx3x3.
        """
        vijs, viis = self._estimate_relative_viewing_directions()

        logger.info("Performing global handedness synchronization.")
        vijs, viis = self._global_J_sync(vijs, viis)

        logger.info("Estimating third rows of rotation matrices.")
        vis = self._estimate_third_rows(vijs, viis)

        logger.info("Estimating in-plane rotations and rotations matrices.")
        Ris = self._estimate_inplane_rotations(vis)

        self.rotations = Ris

    ###########################################
    # Primary Methods                         #
    ###########################################

    def _estimate_relative_viewing_directions(self):
        """
        Estimate the relative viewing directions vij = vi*vj^T, i<j, and vii = vi*vi^T, where
        vi is the third row of the i'th rotation matrix Ri.
        """
        logger.info(f"Estimating relative viewing directions for {self.n_img} images.")
        # Step 1: Detect a single pair of common-lines between each pair of images
        self.build_clmatrix()
        clmatrix = self.clmatrix

        # Step 2: Detect self-common-lines in each image
        sclmatrix = self._self_clmatrix_c3_c4()

        # Step 3: Calculate self-relative-rotations
        Riis = self._estimate_all_Riis_c3_c4(sclmatrix)

        # Step 4: Calculate relative rotations
        Rijs = self._estimate_all_Rijs_c3_c4(clmatrix)

        # Step 5: Inner J-synchronization
        vijs, viis = self._local_J_sync_c3_c4(Rijs, Riis)

        return vijs, viis

    def _global_J_sync(self, vijs, viis):
        """
        Global J-synchronization of all third row outer products. Given 3x3 matrices vijs and viis, each
        of which might contain a spurious J (ie. vij = J*vi*vj^T*J instead of vij = vi*vj^T),
        we return vijs and viis that all have either a spurious J or not.

        :param vijs: An (n-choose-2)x3x3 array where each 3x3 slice holds an estimate for the corresponding
        outer-product vi*vj^T between the third rows of the rotation matrices Ri and Rj. Each estimate
        might have a spurious J independently of other estimates.

        :param viis: An n_imgx3x3 array where the i'th slice holds an estimate for the outer product vi*vi^T
        between the third row of matrix Ri and itself. Each estimate might have a spurious J independently
        of other estimates.

        :return: vijs, viis all of which have a spurious J or not.
        """
        n_img = self.n_img

        # Determine relative handedness of vijs.
        sign_ij_J = self._J_sync_power_method(vijs)

        # Synchronize vijs
        for i, sign in enumerate(sign_ij_J):
            if sign == -1:
                vijs[i] = J_conjugate(vijs[i])

        # Synchronize viis
        # We use the fact that if v_ii and v_ij are of the same handedness, then v_ii @ v_ij = v_ij.
        # If they are opposite handed then Jv_iiJ @ v_ij = v_ij. We compare each v_ii against all
        # previously synchronized v_ij to get a consensus on the handedness of v_ii.
        _, pairs_to_linear = all_pairs(n_img, return_map=True)
        for i in range(n_img):
            vii = viis[i]
            vii_J = J_conjugate(vii)
            J_consensus = 0
            for j in range(n_img):
                if j < i:
                    idx = pairs_to_linear[j, i]
                    vji = vijs[idx]

                    err1 = norm(vji @ vii - vji)
                    err2 = norm(vji @ vii_J - vji)

                elif j > i:
                    idx = pairs_to_linear[i, j]
                    vij = vijs[idx]

                    err1 = norm(vii @ vij - vij)
                    err2 = norm(vii_J @ vij - vij)

                else:
                    continue

                # Accumulate J consensus
                if err1 < err2:
                    J_consensus -= 1
                else:
                    J_consensus += 1

            if J_consensus > 0:
                viis[i] = vii_J
        return vijs, viis

    def _estimate_third_rows(self, vijs, viis):
        """
        Find the third row of each rotation matrix given a collection of matrices
        representing the outer products of the third rows from each rotation matrix.

        :param vijs: An (n-choose-2)x3x3 array where each 3x3 slice holds the third rows
        outer product of the rotation matrices Ri and Rj.

        :param viis: An n_imgx3x3 array where the i'th 3x3 slice holds the outer product of
        the third row of Ri with itself.

        :param order: The underlying molecular symmetry.

        :return: vis, An n_imgx3 matrix whose i'th row is the third row of the rotation matrix Ri.
        """

        n_img = self.n_img

        # Build matrix V whose (i,j)-th block of size 3x3 holds the outer product vij
        V = np.zeros((n_img, n_img, 3, 3), dtype=vijs.dtype)

        # All pairs (i,j) where i<j
        pairs = all_pairs(n_img)

        # Populate upper triangle of V with vijs and lower triangle with vjis, where vji = vij^T.
        for idx, (i, j) in enumerate(pairs):
            V[i, j] = vijs[idx]
            V[j, i] = vijs[idx].T

        # Populate diagonal of V with viis
        for i, vii in enumerate(viis):
            V[i, i] = vii

        # Permute axes and reshape to (3 * n_img, 3 * n_img).
        V = np.swapaxes(V, 1, 2).reshape(3 * n_img, 3 * n_img)

        # In a clean setting V is of rank 1 and its eigenvector is the concatenation
        # of the third rows of all rotation matrices.
        # In the noisy setting we use the eigenvector corresponding to the leading eigenvalue
        val, vec = eigh(V)
        lead_idx = np.argmax(val)
        lead_vec = vec[:, lead_idx]

        # We decompose the leading eigenvector and normalize to obtain the third rows, vis.
        vis = lead_vec.reshape((n_img, 3))
        vis /= anorm(vis, axes=(-1,))[:, np.newaxis]

        return vis

    def _estimate_inplane_rotations(self, vis):
        """
        Estimate the rotation matrices for each image by constructing arbitrary rotation matrices
        populated with the given third rows, vis, and then rotating by an appropriate in-plane rotation.

        :param vis: An n_imgx3 array where the i'th row holds the estimate for the third row of
        the i'th rotation matrix.

        :return: Rotation matrices Ris and in-plane rotation matrices R_thetas, both size n_imgx3x3.
        """
        pf = self.pf
        n_img = self.n_img
        n_theta = self.n_theta
        max_shift_1d = self.max_shift
        shift_step = self.shift_step
        order = self.order
        degree_res = self.degree_res

        # Step 1: Construct all rotation matrices Ri_tildes whose third rows are equal to
        # the corresponding third rows vis.
        Ri_tildes = self._complete_third_row_to_rot(vis)

        # Step 2: Construct all in-plane rotation matrices, R_theta_ijs.
        max_angle = (360 // order) * order
        theta_ijs = np.arange(0, max_angle, degree_res) * np.pi / 180
        R_theta_ijs = Rotation.about_axis("z", theta_ijs, dtype=self.dtype).matrices

        # Step 3: Compute the correlation over all shifts.
        # Generate shifts.
        r_max = pf.shape[-1]
        shifts, shift_phases, _ = self._generate_shift_phase_and_filter(
            r_max, max_shift_1d, shift_step
        )
        n_shifts = len(shifts)

        # Q is the n_img x n_img  Hermitian matrix defined by Q = q*q^H,
        # where q = (exp(i*order*theta_0), ..., exp(i*order*theta_{n_img-1}))^H,
        # and theta_i in [0, 2pi/order) is the in-plane rotation angle for the i'th image.
        Q = np.zeros((n_img, n_img), dtype=complex)

        # Reconstruct the full polar Fourier for use in correlation. self.pf only consists of
        # rays in the range [180, 360), with shape (n_img, n_theta//2, n_rad-1).
        pf = PolarFT.half_to_full(pf)

        # Normalize rays.
        pf /= norm(pf, axis=-1)[..., np.newaxis]

        n_pairs = n_img * (n_img - 1) // 2
        with tqdm(total=n_pairs) as pbar:
            idx = 0
            # Note: the ordering of i and j in these loops should not be changed as
            # they correspond to the ordered tuples (i, j), for i<j.
            for i in range(n_img):
                pf_i = pf[i]

                # Generate shifted versions of images.
                pf_i_shifted = np.array(
                    [pf_i * shift_phase for shift_phase in shift_phases]
                )

                Ri_tilde = Ri_tildes[i]

                for j in range(i + 1, n_img):
                    pf_j = pf[j]

                    Rj_tilde = Ri_tildes[j]

                    # Compute all possible rotations between the i'th and j'th images.
                    Us = np.array(
                        [
                            Ri_tilde.T @ R_theta_ij @ Rj_tilde
                            for R_theta_ij in R_theta_ijs
                        ]
                    )

                    # Find the angle between common lines induced by the rotations.
                    c1s = np.array([[-U[1, 2], U[0, 2]] for U in Us])
                    c2s = np.array([[U[2, 1], -U[2, 0]] for U in Us])

                    # Convert from angles to indices.
                    c1s = self.cl_angles_to_ind(c1s, n_theta)
                    c2s = self.cl_angles_to_ind(c2s, n_theta)

                    # Perform correlation, corrs is shape n_shifts x len(theta_ijs).
                    corrs = np.array(
                        [
                            np.dot(pf_i_shift[c1], np.conj(pf_j[c2]))
                            for pf_i_shift in pf_i_shifted
                            for c1, c2 in zip(c1s, c2s)
                        ]
                    )

                    # Reshape to group by shift and symmetric order.
                    corrs = corrs.reshape((n_shifts, order, len(theta_ijs) // order))

                    # For each pair of lines we take the maximum correlation over all shifts.
                    corrs = np.max(np.real(corrs), axis=0)

                    # corrs[i] is the set of correlations for theta_ij in [2pi * i / order, 2pi * (i + 1) / order).
                    # Due to symmetry, each corrs[i] represents correlations over identical pairs of lines.
                    # With that in mind, we average over corrs[i] and find the max correlation.
                    # This produces an index corresponding to theta_ij in the range [0, 2pi/order).
                    corrs = np.mean(np.real(corrs), axis=0)
                    max_idx_corr = np.argmax(corrs)

                    theta_ij = degree_res * max_idx_corr * np.pi / 180

                    Q[i, j] = np.exp(-1j * order * theta_ij)

                    pbar.update()

                    idx += 1

            # Populate the lower triangle and diagonal of Q.
            # Diagonals are 1 since e^{i*0}=1.
            Q += np.conj(Q).T + np.eye(n_img)

            # Q is a rank-1 Hermitian matrix.
            eig_vals, eig_vecs = eigh(Q)
            leading_eig_vec = eig_vecs[:, -1]
            logger.info(
                f"Top 3 eigenvalues of Q (rank-1) are {str(eig_vals[-3:][::-1])}."
            )

            # Calculate R_thetas.
            R_thetas = Rotation.about_axis(
                "z", np.angle(leading_eig_vec ** (1 / order))
            )

            # Form rotation matrices Ris.
            Ris = R_thetas @ Ri_tildes

            return Ris

    #################################################
    # Secondary Methods for computing outer product #
    #################################################

    def _self_clmatrix_c3_c4(self):
        """
        Find the single pair of self-common-lines in each image assuming that the underlying
        symmetry is C3 or C4.
        """
        pf = self.pf
        n_img = self.n_img
        L = self.src.L
        n_theta = self.n_theta
        max_shift_1d = self.max_shift
        shift_step = self.shift_step
        order = self.order

        # The angle between two self-common-lines is constrained by the underlying symmetry
        # of the molecule (See Lemma A.2 in the listed publication for further details).
        # This angle is in the range [60, 180] for C3 symmetry and [90, 180] for C4 symmetry.
        # Since antipodal lines are perfectly correlated we search for common lines in a smaller window.
        # Note: matlab code used [60, 165] for C3 and [90, 160] for C4.
        # We set the upper bound of this window to be within a factor of 1/L of 180.
        if order == 3:
            min_angle_diff = 60 * np.pi / 180
        else:
            min_angle_diff = 90 * np.pi / 180

        res = 2 * (360 // L)
        max_angle_diff = np.pi - res * np.pi / 180

        # We create a mask associated with angle differences that fall in the
        # range [min_angle_diff, max_angle_diff].

        # `theta_full` and `theta_half` are grids of theta values associate with the polar Fourier transforms
        # `pf_full` and `pf`, respectively.
        theta_vals = np.linspace(0, 360, n_theta)
        theta_vals_half = theta_vals[: len(theta_vals) // 2]
        theta_full, theta_half = np.meshgrid(theta_vals, theta_vals_half)

        # `diff` is the unsigned angle differences between all pairs of polar Fourier rays.
        diff = abs(theta_half - theta_full)
        diff[diff > 180] = 360 - diff[diff > 180]
        diff = np.deg2rad(diff)

        # Build mask.
        good_diffs = (min_angle_diff < diff) & (diff < max_angle_diff)

        # Compute the correlation over all shifts.
        # Generate Shifts.
        r_max = pf.shape[-1]
        shifts, shift_phases, _ = self._generate_shift_phase_and_filter(
            r_max, max_shift_1d, shift_step
        )
        n_shifts = len(shifts)

        # Reconstruct the full polar Fourier for use in correlation. self.pf only consists of
        # rays in the range [180, 360), with shape (n_img, n_theta//2, n_rad-1).
        pf_full = PolarFT.half_to_full(pf)

        # The self-common-lines matrix holds two indices per image that represent
        # the two self common-lines in the image.
        sclmatrix = np.zeros((n_img, 2))

        for i in trange(n_img):
            pf_i = pf[i]
            pf_full_i = pf_full[i]

            # Generate shifted versions of images.
            pf_i_shifted = np.array(
                [pf_i * shift_phase for shift_phase in shift_phases]
            )
            pf_i_shifted = np.reshape(pf_i_shifted, (n_shifts * n_theta // 2, r_max))

            # # Normalize each ray.
            pf_full_i /= norm(pf_full_i, axis=1)[..., np.newaxis]
            pf_i_shifted /= norm(pf_i_shifted, axis=1)[..., np.newaxis]

            # Compute correlation.
            corrs = pf_i_shifted @ pf_full_i.T
            corrs = np.reshape(corrs, (n_shifts, n_theta // 2, n_theta))

            # Mask with allowed combinations.
            corrs *= good_diffs[np.newaxis, ...]

            # Find maximum correlation.
            shift, scl1, scl2 = np.unravel_index(np.argmax(np.real(corrs)), corrs.shape)
            sclmatrix[i] = [scl1, scl2]

        return sclmatrix

    def _estimate_all_Riis_c3_c4(self, sclmatrix):
        """
        Compute estimates for the self relative rotations Rii for every rotation matrix Ri.
        """

        order = self.order
        n_theta = self.n_theta

        # Calculate the cosine of angle between self-common-lines.
        cos_diff = np.cos((sclmatrix[:, 1] - sclmatrix[:, 0]) * 2 * np.pi / n_theta)

        # Calculate Euler angles `euler_y2` (gamma_ii in publication) corresponding to Y in ZYZ convention.
        if order == 3:
            # cos_diff should be <= 0.5, but due to discretization that might be violated.
            if np.max(cos_diff) > 0.5:
                bad_diffs = np.count_nonzero(cos_diff > 0.5)
                logger.debug(
                    "cos(angular_diff) should be < 0.5."
                    f"Found {bad_diffs} estimates exceeding 0.5, with maximum {np.max(cos_diff)}."
                    " Setting all bad estimates to 0.5."
                )
                cos_diff = np.minimum(cos_diff, 0.5)
            euler_y2 = np.arccos(cos_diff / (1 - cos_diff))
        else:
            # cos_diff should be <= 0, but due to discretization that might be violated.
            if np.max(cos_diff) > 0:
                bad_diffs = np.count_nonzero(cos_diff > 0)
                logger.debug(
                    "cos(angular_diff) should be < 0."
                    f"Found {bad_diffs} estimates exceeding 0, with maximum {np.max(cos_diff)}."
                    " Setting all bad estimates to 0."
                )
                cos_diff = np.minimum(cos_diff, 0)
            euler_y2 = np.arccos((1 + cos_diff) / (1 - cos_diff))

        # Calculate remaining Euler angles in ZYZ convention.
        # Note: Publication uses ZXZ convention. Using the notation of the
        # publication the Euler parameterization in ZXZ convention is given by
        # (alpha_ii^(1), gamma_ii, -alpha_ii^(n-1) - pi).
        # Converting to ZYZ convention gives us the parameteriztion
        # (alpha_ii^(1) - pi/2, gamma_ii, -alpha_ii^(n-1) - pi/2).
        euler_z1 = sclmatrix[:, 0] * 2 * np.pi / n_theta - np.pi / 2
        euler_z3 = -sclmatrix[:, 1] * 2 * np.pi / n_theta - np.pi / 2

        # Compute Riis from Euler angles.
        angles = np.array((euler_z1, euler_y2, euler_z3), dtype=self.dtype).T
        Riis = Rotation.from_euler(angles, dtype=self.dtype).matrices

        return Riis

    def _estimate_all_Rijs_c3_c4(self, clmatrix):
        """
        Estimate Rijs using the voting method.
        """
        n_img = self.n_img
        n_theta = self.n_theta
        pairs = all_pairs(n_img)
        Rijs = np.zeros((len(pairs), 3, 3))
        for idx, (i, j) in enumerate(pairs):
            Rijs[idx] = self._syncmatrix_ij_vote_3n(
                clmatrix, i, j, np.arange(n_img), n_theta
            )

        return Rijs

    def _syncmatrix_ij_vote_3n(self, clmatrix, i, j, k_list, n_theta):
        """
        Compute the (i,j) rotation block of the synchronization matrix using voting method

        Given the common lines matrix `clmatrix`, a list of images specified in k_list
        and the number of common lines n_theta, find the (i, j) rotation block Rij.
        :param clmatrix: The common lines matrix
        :param i: The i image
        :param j: The j image
        :param k_list: The list of images for the third image for voting algorithm
        :param n_theta: The number of points in the theta direction (common lines)
        :return: The (i,j) rotation block of the synchronization matrix
        """
        good_k = self._vote_ij(clmatrix, n_theta, i, j, k_list)

        rots = self._rotratio_eulerangle_vec(clmatrix, i, j, good_k, n_theta)

        if rots is not None:
            rot_mean = np.mean(rots, 0)

        else:
            # This is for the case that images i and j correspond to the same
            # viewing direction and differ only by in-plane rotation.
            # We set to zero as in the Matlab code.
            rot_mean = np.zeros((3, 3))

        return rot_mean

    def _local_J_sync_c3_c4(self, Rijs, Riis):
        """
        Estimate viis and vijs. In order to estimate vij = vi @ vj.T, it is necessary for Rii, Rjj,
        and Rij to be of the same handedness. We perform a local handedness synchronization and
        set vij = 1/order * sum(Rii^s @ Rij @ Rjj^s) for s = 0, 1, ..., order. To estimate
        vii = vi @ vi.T we set vii = 1/order * sum(Rii^s) for s = 0, 1, ..., order.

        :param Rijs: An n-choose-2x3x3 array of estimates of relative rotations
            (each pair of images induces two estimates).
        :param Riis: A nx3x3 array of estimates of self-relative rotations.
        :return: vijs, viis
        """
        order = self.order
        n_img = self.n_img
        pairs = all_pairs(n_img)

        # Estimate viis from Riis. vii = 1/order * sum(Rii^s) for s = 0, 1, ..., order.
        viis = np.zeros((n_img, 3, 3))
        for i, Rii in enumerate(Riis):
            viis[i] = np.mean(
                [np.linalg.matrix_power(Rii, s) for s in np.arange(order)], axis=0
            )

        # Estimate vijs via local handedness synchronization.
        vijs = np.zeros((len(pairs), 3, 3))

        for idx, (i, j) in enumerate(pairs):
            opts = np.zeros((2, 2, 2, 3, 3))
            scores_rank1 = np.zeros(8)
            Rii = Riis[i]
            Rjj = Riis[j]
            Rij = Rijs[idx]

            # At this stage, the estimates Rii and Rjj are in the sets {Ri.T @ g^si @ Ri, JRi.T @ g^si @ RiJ}
            # and {Rj.T @ g^sj @ Rj, JRj.T @ g^sj @ RjJ}, respectively, where g^si (and g^sj) is a rotation
            # about the axis of symmetry by (2pi * si)/order, with si = 1 or order-1. Additionally, the estimate
            # Rij might have a spurious J.

            # To estimate vij, it is essential that si = sj and all three estimates Rii, Rjj, and Rij have
            # a spurious J, or none do at all. Note: since g.T = g^(order-1), if Rii = Ri.T @ g^1 @ Ri, then
            # Rii.T = Ri.T @ g^(order-1) @ Ri.

            # The estimate vij should be of rank 1 with singular values [1, 0, 0].
            # We test 8 combinations of handedness and rotation by {g, g^(order-1)} for this condition to determine:
            # a. whether to transpose Rii
            # b. whether to J-conjugate Rii
            # c. whether to J-conjugate Rjj
            if order == 3:
                for ii_trans in [0, 1]:
                    for ii_conj in [0, 1]:
                        for jj_conj in [0, 1]:
                            _Rii = Rii.T if ii_trans else Rii
                            _Rii = J_conjugate(_Rii) if ii_conj else _Rii
                            _Rjj = J_conjugate(Rjj) if jj_conj else Rjj
                            opts[ii_trans, ii_conj, jj_conj] = (
                                Rij + _Rii @ Rij @ _Rjj + _Rii.T @ Rij @ _Rjj.T
                            ) / 3

            else:
                for ii_trans in [0, 1]:
                    for ii_conj in [0, 1]:
                        for jj_conj in [0, 1]:
                            _Rii = Rii.T if ii_trans else Rii
                            _Rii = J_conjugate(_Rii) if ii_conj else _Rii
                            _Rjj = J_conjugate(Rjj) if jj_conj else Rjj
                            opts[ii_trans, ii_conj, jj_conj] = (
                                Rij + _Rii @ Rij @ _Rjj
                            ) / 2

            opts = opts.reshape((8, 3, 3))
            svals = svd(opts, compute_uv=False)
            scores_rank1 = anorm(svals - [1, 0, 0], axes=[1])
            min_idx = np.argmin(scores_rank1)

            # Populate vijs with
            vijs[idx] = opts[min_idx]

        return vijs, viis

    #######################################
    # Secondary Methods for Global J Sync #
    #######################################

    def _J_sync_power_method(self, vijs):
        """
        Calculate the leading eigenvector of the J-synchronization matrix
        using the power method.

        As the J-synchronization matrix is of size (n-choose-2)x(n-choose-2), we
        use the power method to compute the eigenvalues and eigenvectors,
        while constructing the matrix on-the-fly.

        :param vijs: (n-choose-2)x3x3 array of estimates of relative orientation matrices.

        :return: An array of length n-choose-2 consisting of 1 or -1, where the sign of the
        i'th entry indicates whether the i'th relative orientation matrix will be J-conjugated.
        """

        # Set power method tolerance and maximum iterations.
        epsilon = self.epsilon
        max_iters = self.max_iters

        # Initialize candidate eigenvectors
        n_vijs = vijs.shape[0]
        vec = randn(n_vijs, seed=self.seed)
        vec = vec / norm(vec)
        residual = 1
        itr = 0

        # Power method iterations
        logger.info(
            "Initiating power method to estimate J-synchronization matrix eigenvector."
        )
        while itr < max_iters and residual > epsilon:
            itr += 1
            vec_new = self._signs_times_v(vijs, vec)
            vec_new = vec_new / norm(vec_new)
            residual = norm(vec_new - vec)
            vec = vec_new
            logger.info(
                f"Iteration {itr}, residual {round(residual, 5)} (target {epsilon})"
            )

        # We need only the signs of the eigenvector
        J_sync = np.sign(vec)

        return J_sync

    def _signs_times_v(self, vijs, vec):
        """
        Multiplication of the J-synchronization matrix by a candidate eigenvector.

        The J-synchronization matrix is a matrix representation of the handedness graph, Gamma, whose set of
        nodes consists of the estimates vijs and whose set of edges consists of the undirected edges between
        all triplets of estimates vij, vjk, and vik, where i<j<k. The weight of an edge is set to +1 if its
        incident nodes agree in handednes and -1 if not.

        The J-synchronization matrix is of size (n-choose-2)x(n-choose-2), where each entry corresponds to
        the relative handedness of vij and vjk. The entry (ij, jk), where ij and jk are retrieved from the
        all_pairs indexing, is 1 if vij and vjk are of the same handedness and -1 if not. All other entries
        (ij, kl) hold a zero.

        Due to the large size of the J-synchronization matrix we construct it on the fly as follows.
        For each triplet of outer products vij, vjk, and vik, the associated elements of the J-synchronization
        matrix are populated with +1 or -1 and multiplied by the corresponding elements of
        the current candidate eigenvector supplied by the power method. The new candidate eigenvector
        is updated for each triplet.

        :param vijs: (n-choose-2)x3x3 array, where each 3x3 slice holds the outer product of vi and vj.

        :param vec: The current candidate eigenvector of length n-choose-2 from the power method.

        :return: New candidate eigenvector of length n-choose-2. The product of the J-sync matrix and vec.
        """

        # All pairs (i,j) and triplets (i,j,k) where i<j<k
        n_img = self.n_img
        triplets = all_triplets(n_img)
        pairs, pairs_to_linear = all_pairs(n_img, return_map=True)

        # There are 4 possible configurations of relative handedness for each triplet (vij, vjk, vik).
        # 'conjugate' expresses which node of the triplet must be conjugated (True) to achieve synchronization.
        conjugate = np.empty((4, 3), bool)
        conjugate[0] = [False, False, False]
        conjugate[1] = [True, False, False]
        conjugate[2] = [False, True, False]
        conjugate[3] = [False, False, True]

        # 'edges' corresponds to whether conjugation agrees between the pairs (vij, vjk), (vjk, vik),
        # and (vik, vij). True if the pairs are in agreement, False otherwise.
        edges = np.empty((4, 3), bool)
        edges[:, 0] = conjugate[:, 0] == conjugate[:, 1]
        edges[:, 1] = conjugate[:, 1] == conjugate[:, 2]
        edges[:, 2] = conjugate[:, 2] == conjugate[:, 0]

        # The corresponding entries in the J-synchronization matrix are +1 if the pair of nodes agree, -1 if not.
        edge_signs = np.where(edges, 1, -1)

        # For each triplet of nodes we apply the 4 configurations of conjugation and determine the
        # relative handedness based on the condition that vij @ vjk - vik = 0 for synchronized nodes.
        # We then construct the corresponding entries of the J-synchronization matrix with 'edge_signs'
        # corresponding to the conjugation configuration producing the smallest residual for the above
        # condition. Finally, we the multiply the 'edge_signs' by the cooresponding entries of 'vec'.
        v = vijs
        new_vec = np.zeros_like(vec)
        for i, j, k in triplets:
            ij = pairs_to_linear[i, j]
            jk = pairs_to_linear[j, k]
            ik = pairs_to_linear[i, k]
            vij, vjk, vik = v[ij], v[jk], v[ik]
            vij_J = J_conjugate(vij)
            vjk_J = J_conjugate(vjk)
            vik_J = J_conjugate(vik)

            conjugated_pairs = np.where(
                conjugate[..., np.newaxis, np.newaxis],
                [vij_J, vjk_J, vik_J],
                [vij, vjk, vik],
            )
            residual = np.stack([norm(x @ y - z) for x, y, z in conjugated_pairs])

            min_residual = np.argmin(residual)

            # Assign edge weights
            s_ij_jk, s_ik_jk, s_ij_ik = edge_signs[min_residual]

            # Update multiplication of signs times vec
            new_vec[ij] += s_ij_jk * vec[jk] + s_ij_ik * vec[ik]
            new_vec[jk] += s_ij_jk * vec[ij] + s_ik_jk * vec[ik]
            new_vec[ik] += s_ij_jk * vec[ij] + s_ik_jk * vec[jk]

        return new_vec

    ###########################################
    # Secondary Methods fo In-Plane Rotations #
    ###########################################
    @staticmethod
    def _complete_third_row_to_rot(r3):
        """
        Construct rotation matrices whose third rows are equal to the given row vectors.
        For vector r3 = [a, b, c], where [a, b, c] != [0, 0, 1], we return the matrix
        with rows r1, r2, r3, given by:

        r1 = 1/sqrt(a^2 + b^2)[b, -a, 0],
        r2 = 1/sqrt(a^2 + b^2)[ac, bc, -(a^2 + b^2)].

        :param r3: A nx3 array where each row vector has norm 1.
        :return: An nx3x3 array of rotation matrices whose third rows are r3.
        """

        # Handle singleton vector.
        singleton = False
        if r3.shape == (3,):
            r3 = np.expand_dims(r3, axis=0)
            singleton = True

        # Initialize output rotation matrices.
        rots = np.zeros((len(r3), 3, 3), dtype=r3.dtype)

        # Populate 3rd rows.
        rots[:, 2] = r3

        # Mask for third rows that do not coincide with the z-axis.
        mask = np.linalg.norm(r3 - [0, 0, 1], axis=1) >= 1e-5

        # If the third row coincides with the z-axis we return the identity matrix.
        rots[~mask] = np.eye(3, dtype=r3.dtype)

        # 'norm_12' is non-zero since r3 does not coincide with the z-axis.
        norm_12 = np.sqrt(r3[mask, 0] ** 2 + r3[mask, 1] ** 2)

        # Populate 1st rows with vector orthogonal to row 3.
        rots[mask, 0, 0] = r3[mask, 1] / norm_12
        rots[mask, 0, 1] = -r3[mask, 0] / norm_12

        # Populate 2nd rows such that r3 = r1 x r2
        rots[mask, 1, 0] = r3[mask, 0] * r3[mask, 2] / norm_12
        rots[mask, 1, 1] = r3[mask, 1] * r3[mask, 2] / norm_12
        rots[mask, 1, 2] = -norm_12

        if singleton:
            rots = rots.reshape(3, 3)

        return rots

    @staticmethod
    def cl_angles_to_ind(cl_angles, n_theta):
        thetas = np.arctan2(cl_angles[:, 1], cl_angles[:, 0])

        # Shift from [-pi,pi] to [0,2*pi).
        thetas = np.mod(thetas, 2 * np.pi)

        # linear scale from [0,2*pi) to [0,n_theta).
        return np.mod(np.round(thetas / (2 * np.pi) * n_theta), n_theta).astype(int)

    @staticmethod
    def g_sync(rots, order, rots_gt):
        """
        Every estimated rotation might be a version of the ground truth rotation
        rotated by g^{s_i}, where s_i = 0, 1, ..., order. This method synchronizes the
        ground truth rotations so that only a single global rotation need be applied
        to all estimates for error analysis.

        :param rots: Estimated rotation matrices
        :param order: The cyclic order asssociated with the symmetry of the underlying molecule.
        :param rots_gt: Ground truth rotation matrices.

        :return: g-synchronized ground truth rotations.
        """
        assert len(rots) == len(
            rots_gt
        ), "Number of estimates not equal to number of references."
        n_img = len(rots)
        dtype = rots.dtype

        rots_symm = cyclic_rotations(order, dtype).matrices

        A_g = np.zeros((n_img, n_img), dtype=complex)

        pairs = all_pairs(n_img)

        for i, j in pairs:
            Ri = rots[i]
            Rj = rots[j]
            Rij = Ri.T @ Rj

            Ri_gt = rots_gt[i]
            Rj_gt = rots_gt[j]

            diffs = np.zeros(order)
            for s, g_s in enumerate(rots_symm):
                Rij_gt = Ri_gt.T @ g_s @ Rj_gt
                diffs[s] = min([norm(Rij - Rij_gt), norm(Rij - J_conjugate(Rij_gt))])

            idx = np.argmin(diffs)

            A_g[i, j] = np.exp(-1j * 2 * np.pi / order * idx)

        # A_g(k,l) is exp(-j(-theta_k+theta_l))
        # Diagonal elements correspond to exp(-i*0) so put 1.
        # This is important only for verification purposes that spectrum is (K,0,0,0...,0).
        A_g += np.conj(A_g).T + np.eye(n_img)

        _, eig_vecs = eigh(A_g)
        leading_eig_vec = eig_vecs[:, -1]

        angles = np.exp(1j * 2 * np.pi / order * np.arange(order))
        rots_gt_sync = np.zeros((n_img, 3, 3), dtype=dtype)

        for i, rot_gt in enumerate(rots_gt):
            # Since the closest ccw or cw rotation are just as good,
            # we take the absolute value of the angle differences.
            angle_dists = np.abs(np.angle(leading_eig_vec[i] / angles))
            power_g_Ri = np.argmin(angle_dists)
            rots_gt_sync[i] = rots_symm[power_g_Ri] @ rot_gt

        return rots_gt_sync



"""
General purpose math functions, mostly geometric in nature.
"""

import math

import numpy as np
from numpy.linalg import norm
from scipy.linalg import svd

from aspire.utils.random import Random


def cart2pol(x, y):
    """
    Convert Cartesian to Polar Coordinates. All input arguments must be the same shape.

    :param x: x-coordinate in Cartesian space
    :param y: y-coordinate in Cartesian space
    :return: A 2-tuple of values:
        theta: angular coordinate/azimuth
        r: radial distance from origin
    """
    return np.arctan2(y, x), np.hypot(x, y)


def cart2sph(x, y, z):
    """
    Transform cartesian coordinates to spherical. All input arguments must be the same shape.

    :param x: X-values of input co-ordinates.
    :param y: Y-values of input co-ordinates.
    :param z: Z-values of input co-ordinates.
    :return: A 3-tuple of values, all of the same shape as the inputs.
        (<azimuth>, <elevation>, <radius>)
        azimuth and elevation are returned in radians.

    This function is equivalent to MATLAB's cart2sph function.
    """
    hxy = np.hypot(x, y)
    r = np.hypot(hxy, z)
    el = np.arctan2(z, hxy)
    az = np.arctan2(y, x)
    return az, el, r


def _mgrid_slice(n, shifted, normalized):
    """
    Util to generate a `slice` representing a 1d linspace
    as expected by `np.mgrid`.

    :param shifted: shifted by half of grid or not when n is even.
    :param normalized: normalize the grid in the range of (-1, 1) or not.
    :return: `slice` to be used by `np.mgrid`.
    """

    num_points = n * 1j
    start = -n // 2 + 1
    end = n // 2

    if shifted and n % 2 == 0:
        start -= 1 / 2
        end -= 1 / 2
    elif n % 2 == 0:
        start -= 1
        end -= 1

    if normalized:
        # Compute the denominator for normalization
        denom = n / 2
        if shifted and n % 2 == 0:
            denom -= 1 / 2

        # Apply the normalization
        start /= denom
        end /= denom

    return slice(start, end, num_points)


def grid_1d(n, shifted=False, normalized=True, dtype=np.float32):
    """
    Generate one dimensional grid.

    :param n: the number of grid points.
    :param shifted: shifted by half of grid or not when n is even.
    :param normalized: normalize the grid in the range of (-1, 1) or not.
    :return: the rectangular and polar coordinates of all grid points.
    """

    r = x = np.mgrid[_mgrid_slice(n, shifted, normalized)].astype(dtype)

    return {"x": x, "r": r}


def grid_2d(n, shifted=False, normalized=True, indexing="yx", dtype=np.float32):
    """
    Generate two dimensional grid.

    :param n: the number of grid points in each dimension.
    :param shifted: shifted by half of grid or not when n is even.
    :param normalized: normalize the grid in the range of (-1, 1) or not.
    :param indexing: 'yx' (C) or 'xy' (F), defaulting to 'yx'.
        See https://numpy.org/doc/stable/reference/generated/numpy.meshgrid.html
    :return: the rectangular and polar coordinates of all grid points.
    """

    grid = _mgrid_slice(n, shifted, normalized)
    y, x = np.mgrid[grid, grid].astype(dtype)
    if indexing == "xy":
        x, y = y, x
    elif indexing != "yx":
        raise RuntimeError(
            f"grid_2d indexing {indexing} not supported." "  Try 'xy' or 'yx'"
        )

    phi, r = cart2pol(x, y)

    return {"x": x, "y": y, "phi": phi, "r": r}


def grid_3d(n, shifted=False, normalized=True, indexing="zyx", dtype=np.float32):
    """
    Generate three dimensional grid.

    :param n: the number of grid points in each dimension.
    :param shifted: shifted by half of grid or not when n is even.
    :param normalized: normalize the grid in the range of (-1, 1) or not.
    :param indexing: 'zyx' (C) or 'xyz' (F), defaulting to 'zyx'.
        See https://numpy.org/doc/stable/reference/generated/numpy.meshgrid.html
    :return: the rectangular and spherical coordinates of all grid points.
    """

    grid = _mgrid_slice(n, shifted, normalized)
    z, y, x = np.mgrid[grid, grid, grid].astype(dtype)

    if indexing == "xyz":
        x, y, z = z, y, x
    elif indexing != "zyx":
        raise RuntimeError(
            f"grid_3d indexing {indexing} not supported." "  Try 'xyz' or 'zyx'"
        )

    phi, theta, r = cart2sph(x, y, z)

    # TODO: Should this theta adjustment be moved inside cart2sph?
    theta = np.pi / 2 - theta

    return {"x": x, "y": y, "z": z, "phi": phi, "theta": theta, "r": r}


def uniform_random_angles(n, seed=None, dtype=np.float32):
    """
    Generate random 3D rotation angles

    :param n: The number of rotation angles to generate
    :param seed: Random integer seed to use. If None, the current random state is used.
    :return: A n-by-3 ndarray of rotation angles
    """
    # Generate random rotation angles, in radians
    with Random(seed):
        angles = np.column_stack(
            (
                np.random.random(n) * 2 * np.pi,
                np.arccos(2 * np.random.random(n) - 1),
                np.random.random(n) * 2 * np.pi,
            )
        )
    return angles.astype(dtype)


def register_rotations(rots, rots_ref):
    """
    Register estimated orientations to reference ones.

    Finds the orthogonal transformation that best aligns the estimated rotations
    to the reference rotations.

    :param rots: The rotations to be aligned in the form of a n-by-3-by-3 array.
    :param rots_ref: The reference rotations to which we would like to align in
        the form of a n-by-3-by-3 array.
    :return: o_mat, optimal orthogonal 3x3 matrix to align the two sets;
        flag, flag==1 then J conjugacy is required and 0 is not.
    """

    assert (
        rots.shape == rots_ref.shape
    ), "Two sets of rotations must have same dimensions."
    K = rots.shape[0]

    # Reflection matrix
    J = np.array([[1, 0, 0], [0, 1, 0], [0, 0, -1]])

    Q1 = np.zeros((3, 3), dtype=rots.dtype)
    Q2 = np.zeros((3, 3), dtype=rots.dtype)

    for k in range(K):
        R = rots[k, :, :]
        Rref = rots_ref[k, :, :]
        Q1 = Q1 + R @ Rref.T
        Q2 = Q2 + (J @ R @ J) @ Rref.T

    # Compute the two possible orthogonal matrices which register the
    # estimated rotations to the true ones.
    Q1 = Q1 / K
    Q2 = Q2 / K

    # We are registering one set of rotations (the estimated ones) to
    # another set of rotations (the true ones). Thus, the transformation
    # matrix between the two sets of rotations should be orthogonal. This
    # matrix is either Q1 if we recover the non-reflected solution, or Q2,
    # if we got the reflected one. In any case, one of them should be
    # orthogonal.

    err1 = norm(Q1 @ Q1.T - np.eye(3, dtype=rots.dtype), ord="fro")
    err2 = norm(Q2 @ Q2.T - np.eye(3, dtype=rots.dtype), ord="fro")

    # In any case, enforce the registering matrix O to be a rotation.
    if err1 < err2:
        # Use Q1 as the registering matrix
        U, _, V = svd(Q1)
        flag = 0
    else:
        # Use Q2 as the registering matrix
        U, _, V = svd(Q2)
        flag = 1

    Q_mat = U @ V

    return Q_mat, flag


def get_aligned_rotations(rots, Q_mat, flag):
    """
    Get aligned rotation matrices to reference ones.

    Calculated aligned rotation matrices from the orthogonal transformation
    that best aligns the estimated rotations to the reference rotations.


    :param rots: The reference rotations to which we would like to align in
        the form of a n-by-3-by-3 array.
    :param Q_mat:  optimal orthogonal 3x3 transformation matrix
    :param flag:  flag==1 then J conjugacy is required and 0 is not
    :return: regrot, aligned rotation matrices
    """

    K = rots.shape[0]

    # Reflection matrix
    J = np.array([[1, 0, 0], [0, 1, 0], [0, 0, -1]])

    regrot = np.zeros_like(rots)
    for k in range(K):
        R = rots[k, :, :]
        if flag == 1:
            R = J @ R @ J
        regrot[k, :, :] = Q_mat.T @ R

    return regrot


def get_rots_mse(rots_reg, rots_ref):
    """
    Calculate MSE between the estimated orientations to reference ones.

    :param rots_reg: The estimated rotations after alignment in the form of
        a n-by-3-by-3 array.
    :param rots_ref: The reference rotations.
    :return: The MSE value between two sets of rotations.
    """
    assert (
        rots_reg.shape == rots_ref.shape
    ), "Two sets of rotations must have same dimensions."
    K = rots_reg.shape[0]

    diff = np.zeros(K)
    mse = 0
    for k in range(K):
        diff[k] = norm(rots_reg[k, :, :] - rots_ref[k, :, :], ord="fro")
        mse += diff[k] ** 2
    mse = mse / K
    return mse


def common_line_from_rots(r1, r2, ell):
    """
    Compute the common line induced by rotation matrices r1 and r2.

    :param r1: The first rotation matrix of 3-by-3 array.
    :param r2: The second rotation matrix of 3-by-3 array.
    :param ell: The total number of common lines.
    :return: The common line indices for both first and second rotations.
    """

    assert r1.dtype == r2.dtype, "Ambiguous dtypes"

    ut = np.dot(r2, r1.T)
    alpha_ij = np.arctan2(ut[2, 0], -ut[2, 1]) + np.pi
    alpha_ji = np.arctan2(ut[0, 2], -ut[1, 2]) + np.pi

    ell_ij = alpha_ij * ell / (2 * np.pi)
    ell_ji = alpha_ji * ell / (2 * np.pi)

    ell_ij = int(np.mod(np.round(ell_ij), ell))
    ell_ji = int(np.mod(np.round(ell_ji), ell))

    return ell_ij, ell_ji


def rots_to_clmatrix(rots, n_theta):
    """
    Compute the common lines matrix induced by all pairs of rotation
    matrices, `rots`, provided.

    :param rots: n_rotsx3x3 array of rotation matrices.
    :param n_theta: Number of theta values for common lines indices.

    :return: n_rots x n_rots common lines matrix.
    """
    n_rots = len(rots)
    cl_matrix = -np.ones((n_rots, n_rots))
    for i in range(n_rots):
        Ri = rots[i]
        Ri3 = Ri[:, 2]
        for j in range(i + 1, n_rots):
            Rj = rots[j]
            Rj3 = Rj[:, 2]
            common_axis = np.cross(Ri3, Rj3) / np.linalg.norm(np.cross(Ri3, Rj3))
            xij = Ri.T @ common_axis
            xji = Rj.T @ common_axis
            theta_ij = np.rad2deg(np.arctan2(xij[1], xij[0])) % 360
            theta_ji = np.rad2deg(np.arctan2(xji[1], xji[0])) % 360

            if theta_ij > 180:
                theta_ij -= 180
                theta_ji -= 180

            cl_matrix[i, j] = round(theta_ij * n_theta / 360)
            cl_matrix[j, i] = round((theta_ji % 360) * n_theta / 360)

    return cl_matrix


def crop_pad_2d(im, size, fill_value=0):
    """
    :param im: A 2-dimensional numpy array
    :param size: Integer size of cropped/padded output
    :return: A numpy array of shape (size, size)
    """

    im_y, im_x = im.shape
    # shift terms
    start_x = math.floor(im_x / 2) - math.floor(size / 2)
    start_y = math.floor(im_y / 2) - math.floor(size / 2)

    # cropping
    if size <= min(im_y, im_x):
        return im[start_y : start_y + size, start_x : start_x + size]
    # padding
    elif size >= max(im_y, im_x):
        # ensure that we return in the same dtype as the input
        to_return = fill_value * np.ones((size, size), dtype=im.dtype)
        # when padding, start_x and start_y are negative since size is larger
        # than im_x and im_y; the below line calculates where the original image
        # is placed in relation to the (now-larger) box size
        to_return[-start_y : im_y - start_y, -start_x : im_x - start_x] = im
        return to_return
    else:
        # target size is between mat_x and mat_y
        raise ValueError("Cannot crop and pad an image at the same time.")


def crop_pad_3d(im, size, fill_value=0):
    im_y, im_x, im_z = im.shape
    # shift terms
    start_x = math.floor(im_x / 2) - math.floor(size / 2)
    start_y = math.floor(im_y / 2) - math.floor(size / 2)
    start_z = math.floor(im_z / 2) - math.floor(size / 2)

    # cropping
    if size <= min(im_y, im_x, im_z):
        return im[
            start_y : start_y + size, start_x : start_x + size, start_z : start_z + size
        ]
    # padding
    elif size >= max(im_y, im_x, im_z):
        to_return = fill_value * np.ones((size, size, size), dtype=im.dtype)
        to_return[
            -start_y : im_y - start_y,
            -start_x : im_x - start_x,
            -start_z : im_z - start_z,
        ] = im
        return to_return
    else:
        # target size is between min and max of (im_y, im_x, im_z)
        raise ValueError("Cannot crop and pad a volume at the same time.")



import logging

import click

from aspire.basis import FFBBasis2D
from aspire.commands import log_level_option
from aspire.denoising.denoiser_cov2d import DenoiserCov2D
from aspire.noise import AnisotropicNoiseEstimator, WhiteNoiseEstimator
from aspire.source.relion import RelionSource
from aspire.utils.logging import setConsoleLoggingLevel

logger = logging.getLogger(__name__)


@click.command()
@click.option("--data_folder", default=None, help="Path to data folder")
@click.option(
    "--starfile_in",
    required=True,
    help="Path to input starfile relative to project folder",
)
@click.option(
    "--starfile_out",
    required=True,
    help="Path to output starfile relative to project folder",
)
@click.option(
    "--pixel_size", default=1, type=float, help="Pixel size of images in starfile"
)
@click.option(
    "--max_rows",
    default=None,
    type=int,
    help="Max. no. of image rows to read from starfile",
)
@click.option(
    "--max_resolution",
    default=16,
    type=int,
    help="Resolution of downsampled images read from starfile",
)
@click.option(
    "--noise_type", default="White", type=str, help="Noise type for estimation"
)
@click.option(
    "--denoise_method",
    default="CWF",
    type=str,
    help="Specified method for denoising 2D images",
)
@log_level_option
def denoise(
    data_folder,
    starfile_in,
    starfile_out,
    pixel_size,
    max_rows,
    max_resolution,
    noise_type,
    denoise_method,
    loglevel,
):
    """
    Denoise the images and output the clean images using the default CWF method.
    """
    # Set desired logging option for the command line
    setConsoleLoggingLevel(loglevel)

    # Create a source object for 2D images
    logger.info(f"Read in images from {starfile_in} and preprocess the images.")
    source = RelionSource(
        starfile_in, data_folder, pixel_size=pixel_size, max_rows=max_rows
    )

    logger.info(f"Set the resolution to {max_resolution} X {max_resolution}")
    if max_resolution < source.L:
        # Downsample the images
        source = source.downsample(max_resolution)
    else:
        logger.warn(f"Unable to downsample to {max_resolution}, using {source.L}")
    source = source.cache()

    # Specify the fast FB basis method for expending the 2D images
    basis = FFBBasis2D((source.L, source.L))

    # Estimate the noise of images
    noise_estimator = None
    if noise_type == "White":
        logger.info("Estimate the noise of images using white noise method")
        noise_estimator = WhiteNoiseEstimator(source)
    elif noise_type == "Anisotropic":
        logger.info("Estimate the noise of images using anisotropic method")
        noise_estimator = AnisotropicNoiseEstimator(source)
    else:
        raise RuntimeError(f"Unsupported noise_type={noise_type}")

    # Whiten the noise of images
    logger.info("Whiten the noise of images from the noise estimator")
    source = source.whiten(noise_estimator)

    if denoise_method == "CWF":
        logger.info("Denoise the images using CWF cov2D method.")
        denoiser = DenoiserCov2D(source, basis)
        denoised_src = denoiser.denoise(batch_size=512)
        denoised_src.save(
            starfile_out, batch_size=512, save_mode="single", overwrite=False
        )
    else:
        raise NotImplementedError(
            "Currently only covariance Wiener filtering method is supported"
        )



import logging
from functools import partial

import numpy as np
from scipy.linalg import norm
from scipy.sparse.linalg import LinearOperator, cg

from aspire import config
from aspire.nufft import anufft
from aspire.numeric import fft
from aspire.operators import evaluate_src_filters_on_grid
from aspire.reconstruction import Estimator, FourierKernel, FourierKernelMatrix
from aspire.volume import Volume, rotated_grids

logger = logging.getLogger(__name__)


class WeightedVolumesEstimator(Estimator):
    def __init__(self, weights, *args, **kwargs):
        """
        Weighted mean volume estimation.

        This class holds the `FourierKernelMatrix`, stored as a r x r
        matrix of volumes.  The problem being solved here is the
        minimization given by eq. (14) in the paper, rewritten as the
        normal equations in eq. (20) and more compactly in eq. (23).

        Convolution with each of these kernels is equivalent to
        performing a projection/backprojection on a volume, with the
        appropriate amplitude modifiers and CTF, and also a weighting
        term; the r^2 volumes are each of pairwise products between
        the weighting vectors given by the columns of wts.

        Note that this is a non-centered Fourier transform, so the
        zero frequency is found at index 0.

        Formulas and conventions used for Volume estimation are
        described in:

        “Structural Variability from Noisy Tomographic Projections.”
        Andén, Joakim, and Amit Singer.
        SIAM journal on imaging sciences vol. 11,2 (2018): 1441-1492.
        doi:10.1137/17M1153509

        "Cryo-EM reconstruction of continuous heterogeneity by Laplacian spectral volumes"
        Amit Moscovich, Amit Halevi, Joakim Andén and Amit Singer
        Inverse Problems, Volume 36, Number 2, 2020 IOP Publishing Ltd
        Special Issue on Cryo-Electron Microscopy and Inverse Problems
        https://doi.org/10.1088/1361-6420/ab4f55

        :param weights: Matrix of weights, n x r.
        """

        self.weights = weights
        self.r = self.weights.shape[1]
        super().__init__(*args, **kwargs)
        assert self.src.n == self.weights.shape[0]

    def __getattr__(self, name):
        if name == "precond_kernel":
            if self.preconditioner == "circulant":
                self.precond_kernel = FourierKernelMatrix(
                    1.0 / self.kernel.circularize()
                )
            else:
                if self.preconditioner.lower() not in (None, "none"):
                    logger.warning(
                        f"Preconditioner {self.preconditioner} is not implemented, resetting to default of None."
                    )
                self.precond_kernel = None
            return self.precond_kernel

        else:
            #  raise AttributeError(name)
            return super().__getattr__(name)

    def compute_kernel(self):
        """
        Compute and return FourierKernelMatrix instance.
        """
        return FourierKernelMatrix(self._compute_kernel())

    def _compute_kernel(self):
        """
        :return: r x r matrix of volumes, shaped (r, r, 2L, 2L, 2L).
        """

        _2L = 2 * self.src.L
        # Note, because we're iteratively summing it is critical we zero this array.
        kernel = np.zeros((self.r, self.r, _2L, _2L, _2L), dtype=self.dtype)

        for i in range(0, self.src.n, self.batch_size):
            _range = np.arange(i, min(self.src.n, i + self.batch_size), dtype=int)

            pts_rot = rotated_grids(self.src.L, self.src.rotations[_range, :, :])
            pts_rot = pts_rot.reshape((3, -1))
            assert pts_rot.dtype == self.dtype

            sq_filters_f = evaluate_src_filters_on_grid(self.src, _range) ** 2

            amplitudes_sq = (self.src.amplitudes[_range] ** 2).astype(
                self.dtype, copy=False
            )

            for k in range(self.r):
                for j in range(k + 1):
                    weights = sq_filters_f * amplitudes_sq

                    if self.src.L % 2 == 0:
                        weights[0, :, :] = 0
                        weights[:, 0, :] = 0

                    weights *= (
                        self.weights[_range, j] * self.weights[_range, k]
                    ).reshape(1, 1, len(_range))

                    weights = np.transpose(weights, (2, 0, 1)).flatten()

                    batch_kernel = (
                        1
                        / (self.r * self.src.L**4)
                        * anufft(weights, pts_rot[::-1], (_2L, _2L, _2L), real=True)
                    )
                    kernel[k, j] += batch_kernel

                    # r x r symmetric
                    # accumulate batch entries of kernel[k,j] to kernel[j,k]
                    if j != k:
                        kernel[j, k] += batch_kernel

        kermat_f = np.zeros((self.r, self.r, _2L, _2L, _2L))
        logger.info("Computing non-centered Fourier Transform Kernel Mat")
        for k in range(self.r):
            for j in range(self.r):
                # Ensure symmetric kernel
                kernel[k, j, 0, :, :] = 0
                kernel[k, j, :, 0, :] = 0
                kernel[k, j, :, :, 0] = 0

                kernel[k, j] = fft.mdim_ifftshift(kernel[k, j], range(0, 3))
                kernel_f = fft.fftn(kernel[k, j], axes=(0, 1, 2))

                kernel_f = np.real(kernel_f)
                kermat_f[k, j] = kernel_f

        return kermat_f

    def src_backward(self):
        """
        Apply adjoint mapping to source

        :return: The adjoint mapping applied to the images, averaged over the whole dataset and expressed
            as coefficients of `basis`.
        """

        # src_vols_wt_backward
        vol_rhs = Volume(
            np.zeros((self.r, self.src.L, self.src.L, self.src.L), dtype=self.dtype)
        )

        for i in range(0, self.src.n, self.batch_size):
            for k in range(self.r):
                im = self.src.images[i : i + self.batch_size]

                batch_vol_rhs = (
                    self.src.im_backward(im, i, self.weights[:, k]) / self.src.n
                )
                vol_rhs[k] += batch_vol_rhs.astype(self.dtype)

        res = np.sqrt(self.src.n) * self.basis.evaluate_t(vol_rhs)
        logger.info(f"Determined weighted adjoint mappings. Shape = {res.shape}")

        return res

    def conj_grad(self, b_coeff, tol=1e-5, regularizer=0):
        count = b_coeff.shape[-1]  # b_coef should be (r, basis.count)
        kernel = self.kernel

        if regularizer > 0:
            kernel += regularizer

        operator = LinearOperator(
            (self.r * count, self.r * count),
            matvec=partial(self.apply_kernel, kernel=kernel),
            dtype=self.dtype,
        )
        if self.precond_kernel is None:
            M = None
        else:
            precond_kernel = self.precond_kernel
            if regularizer > 0:
                precond_kernel += regularizer
            M = LinearOperator(
                (self.r * count, self.r * count),
                matvec=partial(self.apply_kernel, kernel=precond_kernel),
                dtype=self.dtype,
            )

        tol = tol or config.mean.cg_tol
        target_residual = tol * norm(b_coeff)

        def cb(xk):
            logger.info(
                f"Delta {norm(b_coeff - self.apply_kernel(xk))} (target {target_residual})"
            )

        x, info = cg(operator, b_coeff.flatten(), M=M, callback=cb, tol=tol, atol=0)

        if info != 0:
            raise RuntimeError("Unable to converge!")

        return x.reshape(self.r, self.basis.count)

    def apply_kernel(self, vol_coeff, kernel=None):
        """
        Applies the kernel represented by convolution

        :param vol_coeff: The volume to be convolved, stored in the basis coefficients.
        :param kernel: a Kernel object. If None, the kernel for this Estimator is used.
        :return: The result of evaluating `vol_coeff` in the given basis, convolving with the kernel given by
            kernel, and backprojecting into the basis.
        """
        if kernel is None:
            kernel = self.kernel

        assert np.size(vol_coeff) == self.r * self.basis.count
        if vol_coeff.ndim == 1:
            vol_coeff = vol_coeff.reshape(self.r, self.basis.count)

        vols_out = Volume(
            np.zeros((self.r, self.src.L, self.src.L, self.src.L), dtype=self.dtype)
        )

        vol = self.basis.evaluate(vol_coeff)

        for k in range(self.r):
            for j in range(self.r):
                vols_out[k] = vols_out[k] + kernel.convolve_volume(vol[j], j, k)
                # Note this is where we would add mask_gamma

        vol_coeff = self.basis.evaluate_t(vols_out)

        return vol_coeff


class MeanEstimator(WeightedVolumesEstimator):
    """
    Special case of weighted mean volume estimate,
    for a single volume.
    """

    def __init__(self, src, basis, **kwargs):
        weights = np.ones((src.n, 1)) / np.sqrt(src.n)
        super().__init__(weights, src, basis, **kwargs)

    def __getattr__(self, name):
        """
        See `Estimator.__getattr__`.
        """
        return super(WeightedVolumesEstimator, self).__getattr__(name)

    def apply_kernel(self, *args, **kwargs):
        """
        See `Estimator.apply_kernel`.
        """
        return super(WeightedVolumesEstimator, self).apply_kernel(*args, **kwargs)

    def compute_kernel(self):
        """
        Compute and return `FourierKernel` instance.
        """
        # Note for the r=1 we select and return a single kernel.
        return FourierKernel(self._compute_kernel()[0][0])



import abc
import logging

import numpy as np

from aspire.image import Image
from aspire.image.xform import Xform
from aspire.numeric import fft, xp
from aspire.operators import (
    ArrayFilter,
    BlueFilter,
    PinkFilter,
    PowerFilter,
    ScalarFilter,
)
from aspire.utils import grid_2d, randn, trange

logger = logging.getLogger(__name__)


# TODO: Implement correct hierarchy and DRY


class NoiseAdder(Xform):
    """
    Defines interface for `CustomNoiseAdder`s.
    """

    def __init__(self, noise_filter, seed=0):
        """
        Initialize the random state of this `NoiseAdder` using `noise_filter` and `seed`.

        `noise_filter` will be provided by the user or instantiated automatically by the subclass.

        :param seed: The random seed used to generate white noise.
        :param noise_filter: An `aspire.operators.Filter` object.
            `NoiseAdders` start by generating gaussian noise,
            then apply `noise_filter` to transform the noise.
            Note the `noise_filter` will be raised to the 1/2 power.
        """
        super().__init__()
        self.seed = seed
        self._noise_filter = noise_filter
        self.noise_filter = PowerFilter(noise_filter, power=0.5)

    def __repr__(self):
        return f"{self.__class__.__name__}(noise_filter={self._noise_filter}, seed={self.seed})"

    def __str__(self):
        return f"{self.__class__.__name__}"

    def _forward(self, im, indices):
        _im = im.asnumpy().copy()

        for i, idx in enumerate(indices):
            # Note: The following random seed behavior is directly taken from MATLAB Cov3D code.
            random_seed = self.seed + 191 * (idx + 1)
            im_s = randn(2 * im.resolution, 2 * im.resolution, seed=random_seed)
            # Use numpy because im_s and im are different image sizes
            im_s = Image(im_s).filter(self.noise_filter).asnumpy()[0]
            _im[i] += im_s[: im.resolution, : im.resolution]

        return Image(_im)

    @abc.abstractproperty
    def noise_var(self):
        """
        Concrete implementations are expected to provide a method that
        returns the noise variance for the NoiseAdder.

        Authors of `NoiseAdder`s are encouraged to consider any relevant
        methods of calculating noise variance from theory.
        """


class CustomNoiseAdder(NoiseAdder):
    """
    Instantiates a NoiseAdder using the provided `noise_filter`.
    """

    @property
    def noise_var(self, res=512):
        """
        Return noise variance.

        CustomNoiseAdder will estimate noise_var using the `noise_filter`.

        :param res: Resolution to use when evaluating noise filter, default 512.
        :returns: Noise variance estimated at `res`.
        """
        # Take mean of user provided _noise_filter, before the PowerFilter is applied.
        return np.mean(self._noise_filter.evaluate_grid(res))


class WhiteNoiseAdder(NoiseAdder):
    """
    A Xform that adds white noise, optionally passed through a Filter
    object, to all incoming images.
    """

    # TODO, check if we can change seed and/or why not.
    def __init__(self, var, seed=0):
        """
        Return a `WhiteNoiseAdder` instance from `var` and using `seed`.

        :param var: Target noise variance.
        :param seed: Optinally provide a random seed used to generate white noise.
        """

        self.signal_power = None  # Used with `from_snr`
        self.requires_signal_power = False  # Used with `from_snr`
        self.noise_var = var
        self.seed = seed
        # When we know the var, complete building the filter.
        if var is not None:
            self._build()
        else:
            # Otherwise, we will flag that we require signal power.
            # Assigning `signal_power` later will complete builing
            # filter.
            self.requires_signal_power = True

    def _build(self):
        """
        Builds underlying Filter for this NoiseAdder.
        """
        super().__init__(
            noise_filter=ScalarFilter(dim=2, value=self.noise_var), seed=self.seed
        )

    @classmethod
    def from_snr(cls, snr, signal_power=None, seed=0):
        """
        Generates a WhiteNoiseAdder configured to produce a target
        signal to noise ratio.

        When `signal_power` is not provided, `requires_signal_power`
        attribute will be set.  Consumers can check this attribute and
        set `signal_power` as required. Setting `signal_power` should
        then complete building the filter.

        :param snr: Desired signal to noise ratio of
            the returned source.
        :param signal_power: Optional, if the signal power is known.
        :param seed: Optionally provide a random seed used to generate white noise.
        """

        noise_adder = cls(var=None, seed=seed)
        # signal_power.setter will use `_snr` to compute the noise
        # variance.
        noise_adder._snr = snr

        # `signal_power.setter` should complete _build when provided
        # `signal_power` is not None
        noise_adder.signal_power = signal_power

        return noise_adder

    def __str__(self):
        return f"{self.__class__.__name__} with variance={self._noise_var}"

    def __repr__(self):
        return f"{self.__class__.__name__}(var={self._noise_var}, seed={self.seed})"

    @property
    def noise_var(self):
        """
        Returns noise variance.

        Note in this white noise case, noise variance is known,
        because the `WhiteNoiseAdder` was instantied with an explicit variance.
        """
        return self._noise_var

    @noise_var.setter
    def noise_var(self, v):
        self._noise_var = v

    @property
    def signal_power(self):
        return self._signal_power

    @signal_power.setter
    def signal_power(self, p):
        self._signal_power = p
        if p is not None:
            self.requires_signal_power = False
            self.noise_var = p / self._snr
            self._build()


class BlueNoiseAdder(WhiteNoiseAdder):
    """
    NoiseAdder where noise power increases with frequency.
    """

    def _build(self):
        """
        Builds underlying Filter for this NoiseAdder.
        """

        # Call the __init__ from parent of WhiteNoiseAdder.
        super(WhiteNoiseAdder, self).__init__(
            noise_filter=BlueFilter(var=self.noise_var), seed=self.seed
        )


class PinkNoiseAdder(WhiteNoiseAdder):
    """
    NoiseAdder where noise power decreases with frequency.
    """

    def _build(self):
        """
        Builds underlying Filter for this NoiseAdder.
        """

        # Call the __init__ from parent of WhiteNoiseAdder.
        super(WhiteNoiseAdder, self).__init__(
            noise_filter=PinkFilter(var=self.noise_var), seed=self.seed
        )


class NoiseEstimator:
    """
    Noise Estimator base class.
    """

    def __init__(self, src, bgRadius=1, batchSize=512):
        """
        Any additional args/kwargs are passed on to the Source's 'images' method

        :param src: A Source object which can give us images on demand
        :param bgRadius: The radius of the disk whose complement is used to estimate the noise.
            Radius is relative proportion, where `1` represents
            the radius of disc inscribing a `(src.L, src.L)` image.
        :param batchSize:  The size of the batches in which to compute the variance estimate.
        """
        self.src = src
        self.dtype = self.src.dtype
        self.bgRadius = bgRadius
        self.batchSize = batchSize

        self.filter = self._create_filter()

    def estimate(self):
        """
        :return: The estimated noise variance of the images.
        """
        raise NotImplementedError("Subclasses implement the `estimate` method.")


class WhiteNoiseEstimator(NoiseEstimator):
    """
    White Noise Estimator.
    """

    def estimate(self):
        """
        :return: The estimated noise variance of the images.
        """
        # WhiteNoiseEstimator.filter is a ScalarFilter,
        #   so we only evaluate for the zero frequencies.
        return self.filter.evaluate(np.zeros((2, 1), dtype=self.dtype)).item()

    def _create_filter(self, noise_variance=None):
        """
        :param noise_variance: Noise variance of images
        :return: The estimated noise power spectral distribution (PSD) of the images in the form of a filter object.
        """
        if noise_variance is None:
            logger.info(f"Determining Noise variance in batches of {self.batchSize}")
            noise_variance = self._estimate_noise_variance()
            logger.info(f"Noise variance = {noise_variance}")
        return ScalarFilter(dim=2, value=noise_variance)

    def _estimate_noise_variance(self):
        """
        Any additional arguments/keyword-arguments are passed on to the Source's 'images' method
        :return: The estimated noise variance of the images in the Source used to create this estimator.
        TODO: How's this initial estimate of variance different from the 'estimate' method?
        """
        # Run estimate using saved parameters
        g2d = grid_2d(self.src.L, indexing="yx", dtype=self.dtype)
        mask = g2d["r"] >= self.bgRadius

        first_moment = 0
        second_moment = 0
        for i in trange(0, self.src.n, self.batchSize):
            images = self.src.images[i : i + self.batchSize].asnumpy()
            images_masked = images * mask

            _denominator = self.src.n * np.sum(mask)
            first_moment += np.sum(images_masked) / _denominator
            second_moment += np.sum(np.abs(images_masked**2)) / _denominator
        return second_moment - first_moment**2


class AnisotropicNoiseEstimator(NoiseEstimator):
    """
    Anisotropic White Noise Estimator.
    """

    def estimate(self):
        """
        :return: The estimated noise variance of the images.
        """

        # AnisotropicNoiseEstimator.filter is an ArrayFilter.
        #   We average the variance over all frequencies,

        return np.mean(self.filter.evaluate_grid(self.src.L))

    def _create_filter(self, noise_psd=None):
        """
        :param noise_psd: Noise PSD of images
        :return: The estimated noise power spectral distribution (PSD) of the images in the form of a filter object.
        """
        if noise_psd is None:
            noise_psd = self.estimate_noise_psd()
        return ArrayFilter(noise_psd)

    def estimate_noise_psd(self):
        """
        :return: The estimated noise variance of the images in the Source used to create this estimator.
        TODO: How's this initial estimate of variance different from the 'estimate' method?
        """
        # Run estimate using saved parameters
        g2d = grid_2d(self.src.L, indexing="yx", dtype=self.dtype)
        mask = g2d["r"] >= self.bgRadius

        mean_est = 0
        noise_psd_est = np.zeros((self.src.L, self.src.L)).astype(self.src.dtype)
        for i in trange(0, self.src.n, self.batchSize):
            images = self.src.images[i : i + self.batchSize].asnumpy()
            images_masked = images * mask

            _denominator = self.src.n * np.sum(mask)
            mean_est += np.sum(images_masked) / _denominator
            im_masked_f = xp.asnumpy(fft.centered_fft2(xp.asarray(images_masked)))
            noise_psd_est += np.sum(np.abs(im_masked_f**2), axis=0) / _denominator

        mid = self.src.L // 2
        noise_psd_est[mid, mid] -= mean_est**2

        return noise_psd_est



import numpy as np

from aspire.operators import BlkDiagMatrix


def filter_to_fb_mat(h_fun, fbasis):
    """
    Convert a nonradial function in k space into a basis representation.

    :param h_fun: The function form in k space.
    :param fbasis: The basis object for expanding.

    :return: a BlkDiagMatrix instance representation using the
        `fbasis` expansion.
    """

    # These form a circular dependence, import locally until time to clean up.
    from aspire.basis import FFBBasis2D
    from aspire.basis.basis_utils import lgwt

    if not isinstance(fbasis, FFBBasis2D):
        raise NotImplementedError("Currently only fast FB method is supported")
    # Set same dimensions as basis object
    n_k = fbasis.n_r
    n_theta = fbasis.n_theta
    radial = fbasis.get_radial()

    # get 2D grid in polar coordinate
    k_vals, wts = lgwt(n_k, 0, 0.5, dtype=fbasis.dtype)
    k, theta = np.meshgrid(
        k_vals, np.arange(n_theta) * 2 * np.pi / (2 * n_theta), indexing="ij"
    )

    # Get function values in polar 2D grid and average out angle contribution
    omegax = k * np.cos(theta)
    omegay = k * np.sin(theta)
    omega = 2 * np.pi * np.vstack((omegax.flatten("C"), omegay.flatten("C")))
    h_vals2d = h_fun(omega).reshape(n_k, n_theta).astype(fbasis.dtype)
    h_vals = np.sum(h_vals2d, axis=1) / n_theta

    # Represent 1D function values in fbasis
    h_fb = BlkDiagMatrix.empty(2 * fbasis.ell_max + 1, dtype=fbasis.dtype)
    ind_ell = 0
    for ell in range(0, fbasis.ell_max + 1):
        k_max = fbasis.k_max[ell]
        rmat = 2 * k_vals.reshape(n_k, 1) * fbasis.r0[ell][0:k_max].T
        fb_vals = np.zeros_like(rmat)
        ind_radial = np.sum(fbasis.k_max[0:ell])
        fb_vals[:, 0:k_max] = radial[ind_radial : ind_radial + k_max].T
        h_fb_vals = fb_vals * h_vals.reshape(n_k, 1)
        h_fb_ell = fb_vals.T @ (
            h_fb_vals * k_vals.reshape(n_k, 1) * wts.reshape(n_k, 1)
        )
        h_fb[ind_ell] = h_fb_ell
        ind_ell += 1
        if ell > 0:
            h_fb[ind_ell] = h_fb[ind_ell - 1]
            ind_ell += 1

    return h_fb
